local util = require "./lua/util"
local undo = require "./lua/undo"

local string = string
local math = math
local table = table

local setmetatable = setmetatable
local pairs = pairs
local DoCall = DoCall
local error = error

local MF_ActiveDrawObject = MF_ActiveDrawObject
local MF_ActiveClearObject = MF_ActiveClearObject
local MF_ActiveGetAttribs = MF_ActiveGetAttribs
local MF_Error = MF_Error

module (...)

-- ----------------------------------------------------------------------------

function ProfileClass ()
	return Profile:new()
end

Profile = {
	numActives = 0,
	lastId = 0,
	actives = {},
}

function Profile:new ()
	obj = {}
	setmetatable (obj, self)
	self.__index = self
	return obj
end
	
function Profile:loadProfile( nObj )
	self.numActives = util.tonumber( nObj )
end
	
function Profile:addActive( id, name, filename, ident )
	id = util.tonumber(id)
	self.numActives = self.numActives + 1
	if (id > self.lastId) then
		self.lastId = id
	end
	
	self.actives[id] = {
		id = id,
		name = name,
		file = filename,
		ident = ident,
		values = {}
	}
end

function Profile:newObject (name, filename, ident)
	self.lastId = self.lastId + 1
	self.numActives = self.numActives + 1
	
	self.actives[self.lastId] = {
		id = self.lastId,
		name = name,
		file = filename,
		ident = ident,
		values = {}
	}
end

function Profile:deleteObject (objid)
	objid = util.tonumber (objid)
	self.actives[objid] = nil
	
	self.numActives = self.numActives - 1
end
		
function Profile:drawIndex()
	local ix = 0
	sactives = util.paramsort( self.actives, 'name' )
	for k,v in pairs(sactives) do
		DoCall( "ActiveDrawIndex", v.id, v.name, v.file, (ix % 2), math.floor(ix / 2) )
		ix = ix + 1
	end
	DoCall( "ActiveEndIndex", ix )
end

function Profile:getAttribs( objid )
	objid = util.tonumber( objid )
	if ( self.actives[objid] == nil ) then
		return 0, "", "", ""
	end
	
	return self.actives[objid].id, self.actives[objid].name, self.actives[objid].file, self.actives[objid].ident
end

function Profile:updateAttribs( objid, name, file, ident )
	objid = util.tonumber( objid )
	if ( self.actives[objid] == nil ) then
		return
	end
	
	self.actives[objid].name = name
	self.actives[objid].file = file
	self.actives[objid].ident = ident
end

function Profile:getValue (objid, valid)
	if (self.actives[objid] == nil) then
		return ""
	end
	
	if (self.actives[objid].values[valid] == nil) then
		return 0
	end
	
	return self.actives[objid].values[valid]
end

function Profile:getValueString (objid)
	if (self.actives[objid] == nil) then
		return ""
	end
	
	local vals = {}
	for k,v in pairs (self.actives[objid].values) do
		table.insert (vals, k .. "=" .. v)
	end
	
	return util.str_join (";", vals)
end

-- String format: id1=v1;id2=v2;id3=v3;...
function Profile:setValueString (objid, str)
	if (self.actives[objid] == nil) then
		return
	end
	
	self.actives[objid].values = {}
	
	local vals = util.str_split (";", str)
	for k,v in pairs (vals) do
		if (v ~= "") then
			local parts = util.str_split ("=", v)
			self.actives[objid].values[parts[1]] = parts[2]
		end
	end
end


-- ----------------------------------------------------------------------------

function ActiveClass ()
	return Active:new()
end

Active = {
	filep = { mime = "ACTI", ver = 4, szBlock = 48, szObj = 132, szInst = 12, szObjV = 10, szInstV = 10 },
	inst = {},
	lastInst = 0,
	coffset = nil
}

function Active:new ()
	obj = {
		undoStack = undo.newUndoStack(),
		class = Profile:new(),
		undoAvail = false,
		redoAvail = false,
	}
	setmetatable (obj, self)
	self.__index = self
	return obj
end
	
function Active:bindProfile( prof )
	--self.class.actives = prof.actives
	--self.class = prof
end

-- Primitives -------------------------
	
function Active:add( objid, dx, dy )
	objid = util.tonumber(objid)
	self.lastInst = self.lastInst + 1
	self.inst[self.lastInst] = {
		instid = self.lastInst,
		objid = objid,
		x = dx,
		y = dy,
		values = {}
	}
	
	-- Log undo
	--local udata = { instid = self.lastInst }
	--self:logUndo ("add", udata, 0)
	
	--setmetatable( self.inst[self.lastInst], setmt )
	self:_draw( self.lastInst, objid, self.class.actives[objid].file, dx, dy, self.class.actives[objid].name, self.class.actives[objid].ident )
end
	
function Active:clear( instid )
	instid = util.tonumber(instid)
	if ( instid > 0 ) then
		-- Log undo
		--local udata = util.clone (self.inst[instid])
		--self:logUndo ("clear", udata, 0)
		
		self.inst[instid] = nil
		self:_clear( instid )
	end
end
	

	
function Active:setPosition( instid, x, y )
	instid = util.tonumber(instid)
	if ( self.inst[instid] ~= nil ) then
		-- Log undo
		--local udata = util.clone (self.inst[instid])
		--self:logUndo ("setpos", udata, 0)
		
		local i = self.inst[instid]
		i.x = x
		i.y = y
	end
end

-- Compound Operations ----------------

function Active:addList (actiList)
	local udata = { }
	
	for k,v in pairs (actiList) do
		self:add (v.objid, v.dx, v.dy)
		
		-- Create undo data strucutres
		local instid = self.lastInst
		table.insert (udata, { k = instid, v = util.clone (self.inst[instid]) } )
	end
	
	-- Log undo action
	self:logUndo ("addMulti", udata, 0)
end

function Active:clearList (idlist)
	local udata = { }
	
	for k,v in pairs (idlist) do
		if (self.inst[v] ~= nil) then
			-- Create undo data structures
			table.insert (udata, { k = v, v = util.clone (self.inst[v]) } )
			
			-- Perform clear on single instance
			self.inst[v] = nil
			self:_clear (v)
		end
	end
	
	-- Log undo action
	self:logUndo ("clearMulti", udata, 0)
end

function Active:moveList (diffx, diffy, idlist)
	local udata = { }
	
	for k,v in pairs (idlist) do
		if (self.inst[v] ~= nil) then
			-- Create undo data structures
			table.insert (udata, { k = v, v = util.clone (self.inst[v]) } )
			
			-- Perform position update
			self.inst[v].x = self.inst[v].x + diffx
			self.inst[v].y = self.inst[v].y + diffy
		end
	end
	
	-- Log undo action
	self:logUndo ("moveMulti", udata, 0)
end
	
-- Rendering Primitives ---------------
	
function Active:_draw( instid, objid, file, dx, dy, name, ident )
	if (self.coffset ~= nil) then
		dx, dy = self.coffset (dx, dy)
	end
	--DoCall( "DrawActive", instid, objid, file, dx, dy, name, ident )
	MF_ActiveDrawObject (instid, objid, file, dx, dy, name, ident)
end
	
function Active:_clear( instid )
	--DoCall( "ActiveClear", instid )
	MF_ActiveClearObject (instid)
end

-- Properties -------------------------

function Active:getAttribs (instid)
	local i = self.inst[instid]
	local objname = self.class.actives[i.objid].name
	return instid, i.objid, objname
end

function Active:getValue (instid, valid)
	if (self.actives[instid] == nil) then
		return ""
	end
	
	if (self.actives[instid].values[valid] == nil) then
		return self.class:getValue (self.actives[instid].objid, valid)
	end
	
	return self.actives[instid].values[valid]
end

function Active:getValueString (instid)
	if (self.inst[instid] == nil) then
		return ""
	end
	
	local vals = {}
	for k,v in pairs (self.inst[instid].values) do
		table.insert (vals, k .. "=" .. v)
	end
	
	return util.str_join (";", vals)
end

function Active:setValueString (instid, str)
	if (self.inst[instid] == nil) then
		return
	end
	
	self.inst[instid].values = {}
	
	-- String format: id1=v1;id2=v2;id3=v3;...
	local vals = util.str_split (";", str)
	for k,v in pairs (vals) do
		if (v ~= "") then
			local parts = util.str_split ("=", v)
			self.inst[instid].values[parts[1]] = parts[2]
		end
	end
end

-- Undo / Redo ------------------------

function Active:logUndo (_name, _data, _type)
	local udata = {
		name = _name,
		data = _data
	}
	
	if (_type == 0) then
		self.undoStack:logData (udata)
		self.undoAvail = true
		self.redoAvail = false
	elseif (_type == 1) then
		self.undoStack:logUndo (udata)
		self.undoAvail = true
	elseif (_type == 2) then
		self.undoStack:logRedo (udata)
		self.redoAvail = true
	end
end

function Active:undo ()
	local udata = self.undoStack:Undo()
	if (udata == nil) then
		return false
	end
	
	if (udata.name == "add") then
		self:_undo_add_obj (udata, 2)
	elseif (udata.name == "addMulti") then
		self:_undo_add_multi_obj (udata, 2)
	elseif (udata.name == "clear") then
		self:_undo_clear_obj (udata, 2)
	elseif (udata.name == "clearMulti") then
		self:_undo_clear_multi_obj (udata, 2)
	elseif (udata.name == "setpos") then
		self:_undo_setPosition (udata, 2)
	elseif (udata.name == "moveMulti") then
		self:_undo_move_multi_obj (udata, 2)
	end
	
	self.undoAvail = self.undoStack:UndoAvailable()
	self.redoAvail = self.undoStack:RedoAvailable()
end

function Active:redo ()
	local rdata = self.undoStack:Redo()
	if (rdata == nil) then
		return false
	end

	if (rdata.name == "add") then
		self:_undo_add_obj (rdata, 1)
	elseif (rdata.name == "addMulti") then
		self:_undo_add_multi_obj (rdata, 1)
	elseif (rdata.name == "clear") then
		self:_undo_clear_obj (rdata, 1)
	elseif (rdata.name == "clearMulti") then
		self:_undo_clear_multi_obj (rdata, 1)
	elseif (rdata.name == "setpos") then
		self:_undo_setPosition (rdata, 1)
	elseif (rdata.name == "moveMulti") then
		self:_undo_move_multi_obj (rdata, 1)
	end
	
	self.undoAvail = self.undoStack:UndoAvailable()
	self.redoAvail = self.undoStack:RedoAvailable()
end

function Active:_undo_add_obj (udata, _type)
	-- Log redo action
	local rdata = util.clone (self.inst[udata.data.instid])
	self:logUndo ("clear", rdata, _type)
	
	self.inst[udata.data.instid] = nil
	self:_clear (udata.data.instid)
end

function Active:_undo_add_multi_obj (udata, _type)
	local rdata = { }
	
	for k, v in pairs (udata.data) do
		-- Create redo data structures
		table.insert (rdata, { k = v.k, v = util.clone (v.v) } )
		
		-- Peform undo on single instance
		self.inst[v.k] = nil
		self:_clear (v.k)
	end
	
	-- Log redo action
	self:logUndo ("clearMulti", rdata, _type)
end

function Active:_undo_clear_obj (udata, _type)
	-- Log redo action
	local rdata = { instid = udata.data.instid }
	self:logUndo ("add", rdata, _type)
	
	self.inst[udata.data.instid] = util.clone (udata.data)
	local objid = udata.data.objid
	self:_draw( udata.data.instid, objid, self.class.actives[objid].file, udata.data.x, udata.data.y, 
				self.class.actives[objid].name )
end

function Active:_undo_clear_multi_obj (udata, _type)
	local rdata = { }
	
	for k, v in pairs (udata.data) do
		-- Create redo data structures
		table.insert (rdata, { k = v.k, v = util.clone (v.v) } )
		
		-- Perform undo on single instance
		self.inst[v.k] = util.clone (v.v)
		local objid = v.v.objid
		self:_draw (v.k, objid, self.class.actives[objid].file, v.v.x, v.v.y, self.class.actives[objid].name)
	end
	
	-- Log redo action
	self:logUndo ("addMulti", rdata, _type)
end

function Active:_undo_setPosition (udata, _type)
	-- Log redo action
	local rdata = util.clone (self.inst[udata.data.instid])
	self:logUndo ("setpos", rdata, _type)
	
	self.inst[udata.data.instid] = util.clone (udata.data)
	self:_clear (udata.data.instid)
	
	local objid = udata.data.objid
	self:_draw( udata.data.instid, objid, self.class.actives[objid].file, udata.data.x, udata.data.y, 
				self.class.actives[objid].name )
end

function Active:_undo_move_multi_obj (udata, _type)
	local rdata = { }
	
	for k, v in pairs (udata.data) do
		-- Create redo data structures
		table.insert (rdata, { k = v.k, v = util.clone (self.inst[v.k]) } )
		
		-- Perform undo on single instance
		self.inst[v.k] = util.clone (v.v)
		self:_clear (v.k)
		
		local objid = v.v.objid
		self:_draw (v.k, objid, self.class.actives[objid].file, v.v.x, v.v.y, self.class.actives[objid].name)
	end
	
	-- Log redo action
	self:logUndo ("moveMulti", rdata, _type)
end

-- Import / Export --------------------
	
function Active:import( data, ver )
	local offObj = util.unu32(string.sub(data, 1, 4)) - 8
	local offInst = util.unu32(string.sub(data, 5, 8)) - 8
	local numObj = util.unu32(string.sub(data, 9, 12))
	local numInst = util.unu32(string.sub(data, 13, 16))
	local szObj = util.unu16(string.sub(data, 17, 18))
	local szInst = util.unu16(string.sub(data, 19, 20))

	self.class = Profile:new()
	
	for iObj = 1, numObj do
		local addrObj = offObj + (iObj - 1) * szObj
		local objId = util.unu32(string.sub(data, addrObj+1, addrObj+4))
		local objName = util.unfixstr(string.sub(data, addrObj+5, addrObj+36))
		local objFile = ""
		local objIdent = ""
		
		if (ver > 1) then
			objFile = util.unfixstr(string.sub(data, addrObj+37, addrObj+100))
		end
		if (ver > 2) then
			objIdent = util.unfixstr(string.sub(data, addrObj+101, addrObj+132))
		end
		
		self.class:addActive (objId, objName, objFile, objIdent)
	end
	
	for iInst = 1, numInst do
		local addrInst = offInst + (iInst - 1) * szInst
		local idxObj = util.unu32(string.sub(data, addrInst+1, addrInst+4))
		local x = util.uns32(string.sub(data, addrInst+5, addrInst+8))
		local y = util.uns32(string.sub(data, addrInst+9, addrInst+12))
		self:add( idxObj, x, y )
		
		if (ver < 4) then
			self.inst[self.lastInst].v1 = util.unu32(string.sub(data, addrInst+13, addrInst+16))
			self.inst[self.lastInst].v2 = util.unu32(string.sub(data, addrInst+17, addrInst+20))
			self.inst[self.lastInst].v3 = util.unu32(string.sub(data, addrInst+21, addrInst+24))
			self.inst[self.lastInst].v4 = util.unu32(string.sub(data, addrInst+25, addrInst+28))
			self.inst[self.lastInst].v5 = util.unu32(string.sub(data, addrInst+29, addrInst+32))
			self.inst[self.lastInst].v6 = util.unu32(string.sub(data, addrInst+33, addrInst+36))
			self.inst[self.lastInst].v7 = util.unu32(string.sub(data, addrInst+37, addrInst+40))
			self.inst[self.lastInst].v8 = util.unu32(string.sub(data, addrInst+41, addrInst+44))
			self.inst[self.lastInst].v9 = util.unu32(string.sub(data, addrInst+45, addrInst+48))
			self.inst[self.lastInst].v10 = util.unu32(string.sub(data, addrInst+49, addrInst+52))
			self.inst[self.lastInst].v11 = util.unu32(string.sub(data, addrInst+53, addrInst+56))
			self.inst[self.lastInst].v12 = util.unu32(string.sub(data, addrInst+57, addrInst+60))
			if (ver > 0) then
				self.inst[self.lastInst].persist = util.unu16(string.sub(data, addrInst+61, addrInst+62))
			end
		end
	end
	
	if (ver > 3) then
		local offObjV = util.unu32(string.sub(data, 21, 24)) - 8
		local offInstV = util.unu32(string.sub(data, 25, 28)) - 8
		local numObjV = util.unu32(string.sub(data, 29, 32))
		local numInstV = util.unu32(string.sub(data, 33, 36))
		local szObjV = util.unu16(string.sub(data, 37, 38))
		local szInstV = util.unu16(string.sub(data, 39, 40))
		
		for iObjV = 1, numObjV do
			local addrObjV = offObjV + (iObjV - 1) * szObjV
			local idxObj = util.unu32(string.sub(data, addrObjV+1, addrObjV+4))
			local localID = util.unu16(string.sub(data, addrObjV+5, addrObjV+6))
			local val = util.uns32(string.sub(data, addrObjV+7, addrObjV+10))
			self.class.actives[idxObj].values[localID] = val
		end
		
		for iInstV = 1, numInstV do
			local addrInstV = offInstV + (iInstV - 1) * szInstV
			local idxInst = util.unu32(string.sub(data, addrInstV+1, addrInstV+4))
			local localID = util.unu16(string.sub(data, addrInstV+5, addrInstV+6))
			local val = util.uns32(string.sub(data, addrInstV+7, addrInstV+10))
			self.inst[idxInst].values[localID] = val
		end
	end
end
	
function Active:export()
	local data = ""
	local dataObj = ""
	local dataInst = ""
	local dataObjV = ""
	local dataInstV = ""
	local idxObj = 0
	local idxInst = 0
	local idxObjV = 0
	local idxInstV = 0
	
	for k,v in pairs( self.class.actives ) do
		dataObj = dataObj .. util.u32(v.id) .. util.fixstr(v.name, 32) .. util.fixstr(v.file, 64) .. 
				  util.fixstr(v.ident, 32)
		idxObj = idxObj + 1
		
		for i,j in pairs (v.values) do
			dataObjV = dataObjV .. util.u32(v.id) .. util.u16(i) .. util.s32(j)
			idxObjV = idxObjV + 1
		end
	end
	
	for k,v in pairs( self.inst ) do
		dataInst = dataInst .. util.u32(v.objid) .. util.s32(v.x) .. util.s32(v.y)
		idxInst = idxInst + 1
		
		for i,j in pairs (v.values) do
			dataInstV = dataInstV .. util.u32(v.instid) .. util.u16(i) .. util.s32(j)
			idxInstV = idxInstV + 1
		end
	end
	
	local offsetObj = self.filep.szBlock
	local offsetInst = offsetObj + self.filep.szObj * idxObj
	local offsetObjV = offsetInst + self.filep.szInst * idxInst
	local offsetInstV = offsetObjV + self.filep.szObjV * idxObjV
	
	data = self.filep.mime .. util.u16(self.filep.ver) .. util.u16(self.filep.szBlock) ..
		   util.u32(offsetObj) .. util.u32(offsetInst) ..
		   util.u32(idxObj) .. util.u32(idxInst) .. util.u16(self.filep.szObj) .. util.u16(self.filep.szInst) ..
		   util.u32(offsetObjV) .. util.u32(offsetInstV) .. util.u32(idxObjV) .. util.u32(idxInstV) ..
		   util.u16(self.filep.szObjV) .. util.u16(self.filep.szInstV)
	
	return data .. dataObj .. dataInst .. dataObjV .. dataInstV
end
	
function Active:reset()
	self.inst = {}
	self.lastInst = 0
end