-- Mapedit --------------------------------------------------------------------
-- General-purpose map/level editor targeted for MMF2.
-- Copyright 2007 Justin Aquadro
-- ----------------------------------------------------------------------------
-- Builder interface for Mapedit executable
-- ----------------------------------------------------------------------------

local util = require "./lua/util"
local undo = require "./lua/undo"
local mio = require "./lua/mio"
local tile = require "./lua/tile"
local frame = require "./lua/frame"
--local zone = require "./lua/zone"
local active = require "./lua/active"
local parallax = require "./lua/parallax"
local stringtable = require "./lua/stringtable"

Frame = frame.FrameClass()
--Zone = zone.ZoneClass()
Collision = tile.CollisionClass()
Background = tile.BackgroundClass()
Foreground = tile.ForegroundClass()
Active = active.ActiveClass()
Parallax = parallax.ParallaxClass()
StringTable = stringtable.StringTableClass()

Collision.frame = Frame
Background.frame = Frame
Foreground.frame = Frame

IOC = mio.IOClass()
IOC:registerClass( Frame ) -- This MUST come first
--IOC:registerClass( Zone ) -- This MUST come second
IOC:registerClass( Collision )
IOC:registerClass( Background )
IOC:registerClass( Foreground )
IOC:registerClass( Active )
IOC:registerClass( Parallax )
IOC:registerClass( StringTable )

-- Import / Export ------------------------------------------------------------

function Import( fname )
	IOC:import( fname )
end

function Export( fname )
	IOC:export( fname )
end

function Reset()
	IOC:reset()
end

-- Editor Operations ----------------------------------------------------------

Editor = {
	BTN = { CANCEL = 2, NO = 5, YES = 9 },
	DEFAULT = { WIDTH = 30, HEIGHT = 30 },
	ui = {
		client_width = 0,
		client_height = 0,
		sb_width = 154, 
		sb_height = 0, 
		client_grid_width = 0,
		client_grid_height = 0,
		toolbar_width = 32,
		scrollbar_width = 16,
		tile_size = 16,
		reserved_space = 6,
		pane_text_size = 20,
		pane_spacing = 4,
		scroll_thresh = 48
	},
	winstate = {
		win_width = 640,
		win_height = 480,
		frame_width = 0,
		frame_height = 0,
		frame_x = 0,
		frame_y = 0,
		hscroll = 0,
		vscroll = 0,
		sscroll = 0,
		xmouse = 0,
		ymouse = 0,
		x_tile = 0,
		y_tile = 0,
		oversize_x = 0,
		oversize_y = 0,
		select_valid = 0
	},
	floating = {
		orig_x = 0,
		orig_y = 0,
		x = 0,
		y = 0,
		w = 0,
		h = 0,
		xdiff = 0,
		ydiff = 0
	},
	selected_objects = {},
	selected_drag = {
		orig_x = 0,
		orig_y = 0,
		xdiff = 0,
		ydiff = 0
	},
	select_box = {
		orig_x = 0,
		orig_y = 0
	},
	grid = {
		lastx = 0,
		lasty = 0,
		lastw = 0,
		lasth = 0
	},
	tile_pane_x = 0,
	tile_pane_y = 0,
	tile_pane_w = 1,
	tile_pane_h = 1,
	dirty_pane = 1,
	scroll_pane = 0,
	current_file = "",
	current_name = "Untitled",
	mode = 1,
	modified = 1,
	undo_avail = 0,
	redo_avail = 0,
	cut_copy = 0,
	paste = 0
}

function ClientResize (nw, nh)
	nw = tonumber (nw)
	nh = tonumber (nh)
	
	Editor.ui.sb_height = nh - Editor.ui.toolbar_width
	Editor.ui.client_grid_width = nw - Editor.ui.sb_width - Editor.ui.scrollbar_width
	Editor.ui.client_grid_height = nh - Editor.ui.toolbar_width - Editor.ui.scrollbar_width
end

function SaveFileAs ()
	--local filename, filefilter = DoCall ("INT_SaveFileDialog")
	local filename, filefilter = INT_SaveFileDialog ()
	if (filename ~= nil and filename ~= "") then
		if (filefilter ~= nil) then
			if (tonumber (filefilter) == 2 and string.find (filename, "%.tlv$") == nil) then
				filename = filename .. ".tlv"
			end
		end
		local stat, err = pcall (function() return IOC:export (filename, Frame.key) end)
		if (stat == false) then
			MF_Error (err)
		end
		
		Editor.current_file = filename
		Editor.current_name = util.basename (filename)
		Editor.modified = 1
		ClearModified()
	end
end

function SaveFile ()
	if (Editor.current_file == "") then
		SaveFileAs()
	else
		local stat, err = pcall (function() return IOC:export (Editor.current_file, Frame.key) end)
		if (stat == false) then
			MF_Error (err)
		end
		ClearModified()
	end
end

function PromptSaveFile ()
	if (Editor.modified == 1) then
		--local save = DoCall ("INT_PromptSaveFile")
		local save = INT_PromptSaveFile ()
		if (save == Editor.BTN.CANCEL) then
			return false
		elseif (save == Editor.BTN.YES) then
			SaveFile()
		end
	end
	return true
end

function EndApplication ()
	if (not PromptSaveFile() ) then
		return "cancel"
	end
	return "exit"
end

function NewFile ()
	if (not PromptSaveFile() ) then
		return
	end

	IOC:reset()	
	--DoCall ("MF_ResetFrame")
	MF_ResetFrame ()
	--DoCall ("SetFrameSize", Editor.DEFAULT.WIDTH, Editor.DEFAULT.HEIGHT)
	MF_SetFrameSize (800, 600, 16, 16) --Editor.DEFAULT.WIDTH, Editor.DEFAULT.HEIGHT)
	
	Frame.frameWidth = Editor.DEFAULT.WIDTH
	Frame.frameHeight = Editor.DEFAULT.HEIGHT

	Editor.current_file = ""
	Editor.current_name = "Untitled"
	ClearModified()
end

function OpenGivenFile (filename)
	if (not PromptSaveFile() ) then
		return
	end
	
	IOC:reset()
	--DoCall ("MF_ResetFrame")
	MF_ResetFrame ()
	
	local stat, err = pcall (function () return IOC:import( filename ) end)
	if (stat == false) then
		MF_Error (err)
	end
	
	Editor.current_file = filename
	Editor.current_name = util.basename (filename)
	
	---DoCall ("MF_SetFileName", "Tile Lattice - " .. Editor.current_name)
	MF_SetFileName ("Tile Lattice - " .. Editor.current_name)
	
	EditorRedrawPane()
	ClearModified()
end

function OpenFile ()
	if (not PromptSaveFile() ) then
		return
	end
	
	--local filename = DoCall ("INT_OpenFileDialog")
	local filename = INT_OpenFileDialog ()
	if (filename == nil or filename == "") then
		return
	end
	
	Editor.modified = 0
	OpenGivenFile (filename)
end

function OpenCommandLine (command)
	local subc = string.gsub (command, "\"(.*)\"[ ]*", "", 1)
	
	if (subc ~= nil and subc ~= "") then
		Editor.modified = 0
		OpenGivenFile (subc)
	end
end

function SetModified ()
	if (Editor.modified == 0) then
		Editor.modified = 1
		--DoCall ("MF_SetFileName", "Tile Lattice - " .. Editor.current_name .. "*")
		MF_SetFileName ("Tile Lattice - " .. Editor.current_name .. "*")
	end
	
	CheckUndoActiveState()
	CheckRedoActiveState()
end

function ClearModified ()
	if (Editor.modified == 1) then
		Editor.modified = 0
		--DoCall ("MF_SetFileName", "Tile Lattice - " .. Editor.current_name)
		MF_SetFileName ("Tile Lattice - " .. Editor.current_name)
	end
end

function CheckUndoActiveState ()
	local undostate = false
	if ( (Editor.mode == 0 and Collision.undoAvail) or
		 (Editor.mode == 1 and Background.undoAvail) or
		 (Editor.mode == 2 and Foreground.undoAvail) or
		 (Editor.mode == 3 and Active.undoAvail) ) then
		undostate = true
	end
	
	if (undostate == true and Editor.undo_avail == 0) then
		Editor.undo_avail = 1
		--DoCall ("MF_SetUndoActive")
		MF_SetUndoActive ()
	elseif (undostate == false and Editor.undo_avail == 1) then
		Editor.undo_avail = 0
		--DoCall ("MF_SetUndoInactive")
		MF_SetUndoInactive ()
	end
end

function CheckRedoActiveState ()
	local redostate = false
	if ( (Editor.mode == 0 and Collision.redoAvail) or
		 (Editor.mode == 1 and Background.redoAvail) or
		 (Editor.mode == 2 and Foreground.redoAvail) or 
		 (Editor.mode == 3 and Active.redoAvail) ) then
		redostate = true
	end

	if (redostate == true and Editor.redo_avail == 0) then
		Editor.redo_avail = 1
		--DoCall ("MF_SetRedoActive")
		MF_SetRedoActive ()
	elseif (redostate == false and Editor.redo_avail == 1) then
		Editor.redo_avail = 0
		--DoCall ("MF_SetRedoInactive")
		MF_SetRedoInactive ()
	end
end

function Undo ()
	if (Editor.mode == 0) then
		Collision:undo()
	elseif (Editor.mode == 1) then
		Background:undo()
	elseif (Editor.mode == 2) then
		Foreground:undo()
	elseif (Editor.mode == 3) then
		Active:undo()
	end
	
	CheckUndoActiveState()
	CheckRedoActiveState()
end

function Redo ()
	if (Editor.mode == 0) then
		Collision:redo()
	elseif (Editor.mode == 1) then
		Background:redo()
	elseif (Editor.mode == 2) then
		Foreground:redo()
	elseif (Editor.mode == 3) then
		Active:redo()
	end
	
	CheckUndoActiveState()
	CheckRedoActiveState()
end

function SetMode (mode)
	Editor.mode = tonumber (mode)
	Editor.dirty_pane = 1
	
	CheckUndoActiveState()
	--CheckRedoActiveState()
end

-- Editor Interface -----------------------------------------------------------

function TileToCoord (xtile, ytile)
	xtile = tonumber (xtile)
	ytile = tonumber (ytile)
	
	if (xtile < 0 or xtile >= Frame.frameWidth or
		ytile < 0 or ytile >= Frame.frameHeight) then
		return -1, -1
	end
	
	local xedge = Editor.ui.sb_width + (Editor.ui.reserved_space + Editor.winstate.oversize_x) * Frame.tileWidth
	local xcoord = xedge + xtile * Frame.tileWidth
	local yedge = Editor.ui.toolbar_width + (Editor.ui.reserved_space + Editor.winstate.oversize_y) * Frame.tileHeight
	local ycoord = yedge + ytile * Frame.tileHeight
	
	return xcoord, ycoord
end

Collision.ttc = TileToCoord
Background.ttc = TileToCoord
Foreground.ttc = TileToCoord
--Zone.ttc = TileToCoord

function CoordToTile (xmouse, ymouse, frame_bound)
	xmouse = tonumber (xmouse)
	ymouse = tonumber (ymouse)
	
	if (frame_bound == nil and (
		xmouse <= Editor.winstate.frame_x + Editor.ui.sb_width or
		xmouse > Editor.winstate.frame_x + Editor.winstate.win_width - Editor.ui.scrollbar_width or
		ymouse <= Editor.winstate.frame_y + Editor.ui.toolbar_width or
		ymouse > Editor.winstate.frame_y + Editor.winstate.win_height - Editor.ui.scrollbar_width) ) then
		--return -1, -1
	end

	local min_vscroll = math.floor (Editor.winstate.win_height / 2)
	local min_hscroll = math.floor (Editor.winstate.win_width / 2)
	
	local hscroll_diff = Editor.winstate.hscroll - min_hscroll
	local vscroll_diff = Editor.winstate.vscroll - min_vscroll
	if (Editor.winstate.oversize_x > 0) then
		hscroll_diff = 0
	end
	if (Editor.winstate.oversize_y > 0) then
		vscroll_diff = 0
	end
	
	local x_tile = math.floor ( (xmouse - (Editor.winstate.frame_x - ( (Editor.winstate.frame_x - (hscroll_diff) ) % Frame.tileWidth) )
									    - Editor.ui.sb_width + (hscroll_diff) ) / Frame.tileWidth)
	
	local y_tile = math.floor ( (ymouse - (Editor.winstate.frame_y - ( (Editor.winstate.frame_y - (vscroll_diff) ) % Frame.tileHeight) )
									    - Editor.ui.toolbar_width + (vscroll_diff) ) / Frame.tileHeight)

	x_tile = math.max (-1, x_tile - Editor.ui.reserved_space - Editor.winstate.oversize_x)
	y_tile = math.max (-1, y_tile - Editor.ui.reserved_space - Editor.winstate.oversize_y)

	if (x_tile >= Frame.frameWidth ) then
		x_tile = -1
	end
	if (y_tile >= Frame.frameHeight ) then
		y_tile = -1
	end
	
	return x_tile, y_tile
end

function CoordAddOffset (x, y)
	x = tonumber (x)
	y = tonumber (y)
	
	x = x + Editor.ui.sb_width + Editor.ui.reserved_space * Frame.tileWidth
	x = x + Editor.winstate.oversize_x * Frame.tileWidth
	y = y + Editor.ui.toolbar_width + Editor.ui.reserved_space * Frame.tileHeight
	y = y + Editor.winstate.oversize_y * Frame.tileHeight
	
	return x, y
end

Active.coffset = CoordAddOffset
	
function CoordSubOffset (x, y)
	x = tonumber (x)
	y = tonumber (y)
	
	x = x - Editor.ui.sb_width - Editor.ui.reserved_space * Frame.tileWidth
	x = x - Editor.winstate.oversize_x * Frame.tileWidth
	y = y - Editor.ui.toolbar_width - Editor.ui.reserved_space * Frame.tileHeight
	y = y - Editor.winstate.oversize_y * Frame.tileHeight
	
	return x, y
end
	
function EditorUpdateCoords (xmouse, ymouse)
	xmouse = tonumber (xmouse)
	ymouse = tonumber (ymouse)
	
	Editor.winstate.xmouse = xmouse
	Editor.winstate.ymouse = ymouse
	
	local x_tile = -1
	local y_tile = -1
	x_tile, y_tile = CoordToTile (xmouse, ymouse)

	Editor.winstate.x_tile = x_tile
	Editor.winstate.y_tile = y_tile
end

function EditorRedrawUI (win_width, win_height, frame_width, frame_height, frame_x, frame_y, hscroll, vscroll, sscroll, xmouse, ymouse)
	win_width = tonumber (win_width)
	win_height = tonumber (win_height)
	frame_width = tonumber (frame_width)
	frame_height = tonumber (frame_height)
	frame_x = tonumber (frame_x)
	frame_y = tonumber (frame_y)
	hscroll = tonumber (hscroll)
	vscroll = tonumber (vscroll)
	sscroll = tonumber (sscroll)
	xmouse = tonumber (xmouse)
	ymouse = tonumber (ymouse)
	
	if (hscroll ~= Editor.winstate.hscroll or
		vscroll ~= Editor.winstate.vscroll or
		sscroll ~= Editor.winstate.sscroll) then
		Editor.scroll_pane = 1
	end
	
	if (win_width ~= Editor.winstate.win_width or
		win_height ~= Editor.winstate.win_height) then
		Editor.dirty_pane = 1
		Editor.scroll_pane = 1
	end
	
	Editor.winstate.win_width = win_width
	Editor.winstate.win_height = win_height
	Editor.winstate.frame_width = frame_width
	Editor.winstate.frame_height = frame_height
	Editor.winstate.frame_x = frame_x
	Editor.winstate.frame_y = frame_y
	Editor.winstate.hscroll = hscroll
	Editor.winstate.vscroll = vscroll
	Editor.winstate.sscroll = sscroll
	Editor.winstate.xmouse = xmouse
	Editor.winstate.ymouse = ymouse
	
	local lv_width = Frame.tileWidth * Frame.frameWidth
	local lv_height = Frame.tileHeight * Frame.frameHeight
	local cli_width = lv_width + Frame.tileWidth * Editor.ui.reserved_space * 2
	local cli_height = lv_height + Frame.tileHeight * Editor.ui.reserved_space * 2
	local req_width = Editor.ui.sb_width + cli_width + Editor.ui.scrollbar_width
	local req_height = Editor.ui.toolbar_width + cli_height + Editor.ui.scrollbar_width
	local real_width = math.max (req_width, win_width)
	local real_height = math.max (req_height, win_height)
	
	if (real_width ~= frame_width or real_height ~= frame_height) then
		--DoCall ("MF_EditorResizeFrame", real_width, real_height)
		MF_EditorResizeFrame (real_width, real_height)
	end
	
	local x_diff = 0
	local y_diff = 0
	if (real_width > req_width) then
		x_diff = (real_width - req_width) / 2
		x_diff = x_diff - (x_diff % Frame.tileWidth)
		Editor.winstate.oversize_x = math.floor (x_diff / Frame.tileWidth)
		--DoCall ("MF_EditorDisableHScroll")
		MF_EditorDisableHScroll ()
	else
		Editor.winstate.oversize_y = 0
		--DoCall ("MF_EditorEnableHScroll")
		MF_EditorEnableHScroll ()
	end
	if (real_height > req_height) then
		y_diff = (real_height - req_height) / 2
		y_diff = y_diff - (y_diff % Frame.tileHeight)
		Editor.winstate.oversize_y = math.floor (y_diff / Frame.tileHeight)
		--DoCall ("MF_EditorDisableVScroll")
		MF_EditorDisableVScroll ()
	else
		Editor.winstate.oversize_y = 0
		--DoCall ("MF_EditorEnableVScroll")
		MF_EditorEnableVScroll ()
	end
	
	local x_lv = Editor.ui.sb_width + Frame.tileWidth * Editor.ui.reserved_space + x_diff
	local y_lv = Editor.ui.toolbar_width + Frame.tileHeight * Editor.ui.reserved_space + y_diff
	--DoCall ("MF_EditorDrawEditSpace", x_lv, y_lv, lv_width, lv_height)
	MF_EditorDrawEditSpace (x_lv, y_lv, lv_width, lv_height)
	
	local x_hscroll = frame_x + Editor.ui.sb_width
	local y_hscroll = frame_y + win_height - Editor.ui.scrollbar_width
	local len_hscroll = win_width - Editor.ui.sb_width - Editor.ui.scrollbar_width
	local r1_hscroll = win_width / 2
	local r2_hscroll = req_width - (win_width / 2)
	--DoCall ("MF_EditorDrawHScroll", x_hscroll, y_hscroll, len_hscroll, r1_hscroll, r2_hscroll)
	MF_EditorDrawHScroll (x_hscroll, y_hscroll, len_hscroll, r1_hscroll, r2_hscroll)
	
	local x_vscroll = frame_x + win_width - Editor.ui.scrollbar_width
	local y_vscroll = frame_y + Editor.ui.toolbar_width
	local len_vscroll = win_height - Editor.ui.toolbar_width - Editor.ui.scrollbar_width
	local r1_vscroll = win_height / 2
	local r2_vscroll = req_height - (win_height / 2)
	--DoCall ("MF_EditorDrawVScroll", x_vscroll, y_vscroll, len_vscroll, r1_vscroll, r2_vscroll)
	MF_EditorDrawVScroll (x_vscroll, y_vscroll, len_vscroll, r1_vscroll, r2_vscroll)
	--DoCall ("MF_EditorDrawScrollJoint", x_vscroll, y_hscroll)
	MF_EditorDrawScrollJoint (x_vscroll, y_hscroll)
	
	local x_grid = frame_x + Editor.ui.sb_width - (frame_x % Frame.tileWidth)
	local y_grid = frame_y + Editor.ui.toolbar_width - (frame_y % Frame.tileHeight)
	local w_grid = frame_width - Editor.ui.sb_width - Editor.ui.scrollbar_width + (frame_width % Frame.tileWidth)
	local h_grid = frame_height - Editor.ui.toolbar_width - Editor.ui.scrollbar_width + (frame_height % Frame.tileHeight)
	--DoCall ("MF_EditorDrawGrid", x_grid, y_grid, w_grid, h_grid)
	--MF_EditorDrawGrid (x_grid, y_grid, w_grid, h_grid)
	
	if (x ~= Editor.grid.lastx or y ~= Editor.grid.lasty or w ~= Editor.grid.lastw or h ~= Editor.grid.lasth) then
		EditorDrawGrid (x_grid, y_grid, w_grid, h_grid)
	end
	
	local x_sscroll = frame_x + Editor.ui.sb_width - Editor.ui.scrollbar_width - 2
	local y_sscroll = frame_y + Editor.ui.toolbar_width + 1
	local len_sscroll = win_height - Editor.ui.toolbar_width - Editor.ui.scrollbar_width - 2
	--DoCall ("MF_EditorDrawSideScroll", x_sscroll, y_sscroll, len_sscroll)
	MF_EditorDrawSideScroll (x_sscroll, y_sscroll, len_sscroll)
	
	local x_sb = frame_x - 1
	local y_sb = frame_y + Editor.ui.toolbar_width
	local w_sb = Editor.ui.sb_width + 1
	local h_sb = win_height - Editor.ui.toolbar_width + 1
	--DoCall ("MF_EditorDrawSidebar", x_sb, y_sb, w_sb, h_sb, frame_x, y_sb + len_sscroll + 1, w_sb - 3)
	MF_EditorDrawSidebar (x_sb, y_sb, w_sb, h_sb, frame_x, y_sb + len_sscroll + 1, w_sb - 3)
	
	local y_status = frame_y + win_height - Editor.ui.scrollbar_width
	--DoCall ("MF_EditorDrawStatus", frame_x, y_status, w_sb - 3, Editor.ui.scrollbar_width)
	MF_EditorDrawStatus (frame_x, y_status, w_sb - 3, Editor.ui.scrollbar_width)
	
	--DoCall ("MF_EditorDrawToolbar", frame_x, frame_y, win_width + 1, Editor.ui.toolbar_width)
	MF_EditorDrawToolbar (frame_x, frame_y, win_width + 1, Editor.ui.toolbar_width)
	
	-- Toolbar
	local x_tbspace = 29
	local x_tb_file = frame_x
	local x_tb_undo = x_tb_file + (x_tbspace * 3)
	local x_tb_tool = x_tb_undo + (x_tbspace * 2) + 3
	local x_tb_mode = x_tb_tool + (x_tbspace * 3) + 3
	local x_tb_misc = x_tb_mode + (x_tbspace * 8) + 3
	
	--DoCall ("MF_EditorToolbarFile", x_tb_file, frame_y + 2, x_tbspace)
	--DoCall ("MF_EditorToolbarUndo", x_tb_undo, frame_y + 2, x_tbspace)
	--DoCall ("MF_EditorToolbarTool", x_tb_tool, frame_y + 2, x_tbspace)
	--DoCall ("MF_EditorToolbarMode", x_tb_mode, frame_y + 2, x_tbspace)
	--DoCall ("MF_EditorToolbarMisc", x_tb_misc, frame_y + 2, x_tbspace)
	
	MF_EditorToolbarFile (x_tb_file, frame_y + 2, x_tbspace)
	MF_EditorToolbarUndo (x_tb_undo, frame_y + 2, x_tbspace)
	MF_EditorToolbarTool (x_tb_tool, frame_y + 2, x_tbspace)
	MF_EditorToolbarMode (x_tb_mode, frame_y + 2, x_tbspace)
	MF_EditorToolbarMisc (x_tb_misc, frame_y + 2, x_tbspace)
	
	if (Editor.dirty_pane == 1) then
		EditorRedrawPane (1)
	end
	if (Editor.scroll_pane == 1) then
		EditorScrollPane (1)
	end
	
	-- Status Area
	if (Editor.winstate.x_tile == -1 or Editor.winstate.y_tile == -1) then
		--DoCall ("MF_EditorSetStatus", "")
		MF_EditorSetStatus ("")
	else
		local pos = ' (' .. Editor.winstate.x_tile .. ', ' .. Editor.winstate.y_tile .. ')'
		--DoCall ("MF_EditorSetStatus", pos)
		MF_EditorSetStatus (pos)
	end
end

function EditorDrawGrid (x, y, w, h)

	MF_EditorClearGrid ()
	
	if (x ~= nil) then
		Editor.grid.lastx = x
		Editor.grid.lasty = y
		Editor.grid.lastw = w
		Editor.grid.lasth = h
	else
		x = Editor.grid.lastx
		y = Editor.grid.lasty
		w = Editor.grid.lastw
		h = Editor.grid.lasth
	end
	
	local x_witer = math.ceil (w / Frame.tileWidth)
	local y_hiter = math.ceil (h / Frame.tileHeight)
	
	for xi = 0, x_witer do
		MF_EditorDrawGridX (x + xi * Frame.tileWidth - 1, y, 1, h)
	end
	
	for yi = 0, y_hiter do
		MF_EditorDrawGridY (x, y + yi * Frame.tileHeight - 1, w, 1)
	end
end

function EditorMouseInFrame ()
	if (Editor.winstate.xmouse < Editor.winstate.frame_x + Editor.ui.sb_width) then
		MF_MouseOutsideFrame ()
		return
	elseif (Editor.winstate.xmouse >= Editor.winstate.frame_x + Editor.winstate.win_width - Editor.ui.scrollbar_width) then
		MF_MouseOutsideFrame ()
		return
	elseif (Editor.winstate.ymouse < Editor.winstate.frame_y + Editor.ui.toolbar_width) then
		MF_MouseOutsideFrame ()
		return
	elseif (Editor.winstate.ymouse >= Editor.winstate.frame_y + Editor.winstate.win_height - Editor.ui.scrollbar_width) then
		MF_MouseOutsideFrame ()
		return
	end
	
	MF_MouseInsideFrame ()
end

function EditorTestPaneScroll (req_scroll)
	local scr_area = Editor.winstate.win_height - Editor.ui.toolbar_width - Editor.ui.scrollbar_width - 2
	if (req_scroll < scr_area) then
		--DoCall ("MF_EditorDisableSScroll")
		MF_EditorDisableSScroll ()
	else
		--DoCall ("MF_EditorEnableSScroll")
		MF_EditorEnableSScroll ()
	end
	
	local range = req_scroll - scr_area
	if (range >= 0) then
		--DoCall ("MF_EditorRangeSScroll", range)
		MF_EditorRangeSScroll (range)
	end
end

function EditorRedrawPane (bit)
	if (Editor.dirty_pane == 0) then
		Editor.dirty_pane = 1
	end
	
	if (bit == nil or bit == '') then
		return
	end
	
	if (Editor.mode == 0) then
		EditorRedrawTilePane ('Collision')
	elseif (Editor.mode == 1) then
		EditorRedrawTilePane ('Background')
	elseif (Editor.mode == 2) then
		EditorRedrawTilePane ('Foreground')
	elseif (Editor.mode == 3) then
		EditorRedrawObjectPane()
	elseif (Editor.mode == 4) then
		EditorRedrawParallaxPane()
	--elseif (Editor.mode == 6) then
	--	EditorRedrawZonePane()
	end
	
	Editor.dirty_pane = 0
	Editor.scroll_pane = 0
end

function EditorScrollPane (bit)
	if (Editor.scroll_pane == 0) then
		Editor.scroll_pane = 1
	end
	
	if (bit == nil or bit == '') then
		return
	end
	
	if (Editor.mode == 0) then
		EditorScrollTilePane ('Collision')
	elseif (Editor.mode == 1) then
		EditorScrollTilePane ('Background')
	elseif (Editor.mode == 2) then
		EditorScrollTilePane ('Foreground')
	elseif (Editor.mode == 3) then
		EditorScrollObjectPane()
	elseif (Editor.mode == 4) then
		EditorScrollParallaxPane()
	--elseif (Editor.mode == 6) then
	--	EditorScrollZonePane()
	end
	
	Editor.scroll_pane = 0
end

function EditorRedrawTilePane (layer)
	Editor.tile_pane_x = 0
	Editor.tile_pane_y = 0
	Editor.tile_pane_w = 1
	Editor.tile_pane_h = 1
	
	local x_tile = Editor.winstate.frame_x
	local y_tile = Editor.winstate.frame_y + Editor.ui.toolbar_width + 1 - Editor.winstate.sscroll
	local sx_tile = Frame.tileWidth + 1
	local sy_tile = Frame.tileHeight + 1
	
	local obj = GetObject (layer)
	if (obj == nil) then
		return
	end
	
	EditorTestPaneScroll (obj.chip_height)
	
	--DoCall ("MF_EditorDraw" .. layer .. "Pane", x_tile, y_tile, s_tile, s_tile)
	if (layer == 'Collision') then
		MF_EditorDrawCollisionPane (x_tile, y_tile, sx_tile * Editor.tile_pane_w, sy_tile * Editor.tile_pane_h)
	elseif (layer == 'Background') then
		MF_EditorDrawBackgroundPane (x_tile, y_tile, sx_tile * Editor.tile_pane_w, sy_tile * Editor.tile_pane_h)
	elseif (layer == 'Foreground') then
		MF_EditorDrawForegroundPane (x_tile, y_tile, sx_tile * Editor.tile_pane_w, sy_tile * Editor.tile_pane_h)
	end
end

function EditorScrollTilePane (layer)
	local x_tile = Editor.winstate.frame_x
	local y_tile = Editor.winstate.frame_y + Editor.ui.toolbar_width + 1 - Editor.winstate.sscroll
	
	local sx = x_tile + (Editor.tile_pane_x * (Frame.tileWidth + 1) )
	local sy = y_tile + (Editor.tile_pane_y * (Frame.tileHeight + 1) )
	local sx_tile = Frame.tileWidth + 1
	local sy_tile = Frame.tileHeight + 1
	
	local obj = GetObject (layer)
	if (obj == nil) then
		return
	end
	
	--DoCall ("MF_EditorScroll" .. layer .. "Pane", x_tile, y_tile)
	if (layer == 'Collision') then
		MF_EditorScrollCollisionPane (x_tile, y_tile, sx, sy, sx_tile * Editor.tile_pane_w, sy_tile * Editor.tile_pane_h)
	elseif (layer == 'Background') then
		MF_EditorScrollBackgroundPane (x_tile, y_tile, sx, sy, sx_tile * Editor.tile_pane_w, sy_tile * Editor.tile_pane_h)
	elseif (layer == 'Foreground') then
		MF_EditorScrollForegroundPane (x_tile, y_tile, sx, sy, sx_tile * Editor.tile_pane_w, sy_tile * Editor.tile_pane_h)
	end
end

function EditorRedrawObjectPane ()
	--DoCall ("MF_EditorClearObjectPane")
	MF_EditorClearObjectPane ()

	local req_scroll = math.ceil (ProfileNumObjects() / 2.0) * 88 - 1
	EditorTestPaneScroll (req_scroll)
	
	local sactives = util.paramsort(Active.class.actives, 'name' )
	local pos_id = 0
	for k,v in pairs (sactives) do
		local s_obj = (Editor.ui.sb_width - Editor.ui.scrollbar_width - Editor.ui.pane_spacing - 6) / 2
		local x_off = Editor.winstate.frame_x + 2 + (pos_id % 2) * (s_obj + Editor.ui.pane_spacing)
		local y_off = Editor.winstate.frame_y + Editor.ui.toolbar_width - Editor.winstate.sscroll + 3 + 
					  math.floor (pos_id / 2.0) * (s_obj + Editor.ui.pane_text_size)
					  
		--DoCall ("MF_ActiveDrawIndex", math.floor (pos_id / 2.0), v.id, v.name, v.file, x_off, y_off, s_obj)
		MF_ActiveDrawIndex (math.floor (pos_id / 2.0), v.id, v.name, v.file, x_off, y_off, s_obj)
		pos_id = pos_id + 1
	end
end

function EditorScrollObjectPane ()
	local sactives = util.paramsort(Active.class.actives, 'name' )
	local pos_id = 0
	for k,v in pairs (sactives) do
		local s_obj = (Editor.ui.sb_width - Editor.ui.scrollbar_width - Editor.ui.pane_spacing - 6) / 2
		local x_off = Editor.winstate.frame_x + 2 + (pos_id % 2) * (s_obj + Editor.ui.pane_spacing)
		local y_off = Editor.winstate.frame_y + Editor.ui.toolbar_width - Editor.winstate.sscroll + 3 + 
					  math.floor (pos_id / 2.0) * (s_obj + Editor.ui.pane_text_size)
					  
		--DoCall ("MF_ActiveScrollIndex", math.floor (pos_id / 2.0), v.id, x_off, y_off, s_obj)
		MF_ActiveScrollIndex (math.floor (pos_id / 2.0), v.id, x_off, y_off, s_obj)
		pos_id = pos_id + 1
	end
end

function EditorRedrawParallaxPane ()
	--DoCall ("MF_EditorClearParallaxPane")
	MF_EditorClearParallaxPane ()

	local req_scroll = Parallax.numBackdrops * 156 - 1
	EditorTestPaneScroll (req_scroll)
	
	ParallaxDrawIndex()
end

function EditorScrollParallaxPane ()
	local bgs = Parallax:zList()
	local pos_id = 0
	for k,v in pairs (bgs) do
		local s_obj = Editor.ui.sb_width - Editor.ui.scrollbar_width - 6
		local x_off = Editor.winstate.frame_x + 2
		local y_off = Editor.winstate.frame_y + Editor.ui.toolbar_width - Editor.winstate.sscroll + 3 +
					  pos_id * (s_obj + Editor.ui.pane_text_size)
		
		--DoCall ("MF_ParallaxScrollIndex", pos_id, v.id, x_off, y_off, s_obj)
		MF_ParallaxScrollIndex (pos_id, v.id, x_off, y_off, s_obj)
		pos_id = pos_id + 1
	end
end

--[[function EditorRedrawZonePane ()
	DoCall( "MF_ZoneClearIndex" )

	local req_scroll = #Zone.zones * 24 - 1
	EditorTestPaneScroll (req_scroll)
	
	ZoneDrawIndex()
end

function EditorScrollZonePane ()
	for i = 1,#Zone.zones do
		local sw_obj = Editor.ui.sb_width - Editor.ui.scrollbar_width - 6 - 24
		local sh_obj = 24
		local x_off = Editor.winstate.frame_x
		local y_off = Editor.winstate.frame_y + Editor.ui.toolbar_width - Editor.winstate.sscroll + 1 +
					  (i - 1) * sh_obj

		DoCall( "MF_ZoneScrollIndex", i, x_off, y_off, sw_obj, sh_obj )
	end
end]]

function EditorSelect (xmouse, ymouse)
	xmouse = tonumber (xmouse)
	ymouse = tonumber (ymouse)
	
	Editor.winstate.select_valid = 0
	
	local rsx = Editor.ui.reserved_space * Frame.tileWidth
	local rsy = Editor.ui.reserved_space * Frame.tileHeight
	local thresh_x1 = Editor.ui.sb_width + rsx + Editor.winstate.oversize_x * Frame.tileWidth
	local thresh_x2 = thresh_x1 + Frame.frameWidth * Frame.tileWidth
	local thresh_y1 = Editor.ui.toolbar_width + rsy + Editor.winstate.oversize_y * Frame.tileHeight
	local thresh_y2 = thresh_y1 + Frame.frameHeight * Frame.tileHeight
	
	thresh_x1 = math.max (thresh_x1, Editor.winstate.frame_x + Editor.ui.sb_width)
	thresh_x2 = math.min (thresh_x2, Editor.winstate.frame_x + Editor.winstate.win_width - Editor.ui.scrollbar_width)
	thresh_y1 = math.max (thresh_y1, Editor.winstate.frame_y + Editor.ui.toolbar_width)
	thresh_y2 = math.min (thresh_y2, Editor.winstate.frame_y + Editor.winstate.win_height - Editor.ui.scrollbar_width)
	
	if (xmouse < thresh_x1 or xmouse >= thresh_x2 or ymouse < thresh_y1 or ymouse >= thresh_y2) then
		--DoCall ("MF_EditorClearSelect")
		MF_EditorClearSelect ()
		return
	end
	
	Editor.winstate.select_valid = 1	
	Editor.select_box.orig_x = xmouse
	Editor.select_box.orig_y = ymouse
	
	--DoCall ("MF_EditorSelect", xmouse, ymouse, 1, 1)
	MF_EditorSelect (xmouse, ymouse, 1, 1)
end

function EditorSnappingSelect (xmouse, ymouse)
	local tile_x
	local tile_y
	tile_x, tile_y = CoordToTile (xmouse, ymouse)
	
	Editor.winstate.select_valid = 0
	
	if (tile_x == -1 or tile_y == -1) then
		--DoCall ("MF_EditorClearSelect")
		MF_EditorClearSelect ()
		return
	end
	
	local coord_x
	local coord_y
	coord_x, coord_y = TileToCoord (tile_x, tile_y)
	
	Editor.winstate.select_valid = 1
	Editor.select_box.orig_x = coord_x
	Editor.select_box.orig_y = coord_y
	
	--[[if ( (xmouse - coord_x) % Frame.tileWidth > (Frame.tileWidth / 2) ) then
		tile_x = math.min (tile_x + 1, Frame.frameWidth)
		coord_x, coord_y = TileToCoord (tile_x, tile_y)
	end
	if ( (ymouse - coord_y) % Frame.tileHeight > (Frame.tileHeight / 2) ) then
		tile_y = math.min (tile_y + 1, Frame.frameHeight)
		coord_x, coord_y = TileToCoord (tile_x, tile_y)
	end]]
	
	--DoCall ("MF_EditorSnappingSelect", coord_x, coord_y, 1, 1)
	MF_EditorSnappingSelect (coord_x, coord_y, 1, 1)
end	

function EditorExpandSelect (xsrc, ysrc, xmouse, ymouse)
	xsrc = tonumber (xsrc)
	ysrc = tonumber (ysrc)
	xmouse = tonumber (xmouse)
	ymouse = tonumber (ymouse)
	
	if (xmouse < Editor.select_box.orig_x) then
		xsrc = Editor.select_box.orig_x
	end
	if (ymouse < Editor.select_box.orig_y) then
		ysrc = Editor.select_box.orig_y
	end
	
	local rsx = Editor.ui.reserved_space * Frame.tileWidth
	local rsy = Editor.ui.reserved_space * Frame.tileHeight
	local thresh_x1 = Editor.ui.sb_width + rsx + Editor.winstate.oversize_x * Frame.tileWidth
	local thresh_x2 = thresh_x1 + Frame.frameWidth * Frame.tileWidth
	local thresh_y1 = Editor.ui.toolbar_width + rsy + Editor.winstate.oversize_y * Frame.tileHeight
	local thresh_y2 = thresh_y1 + Frame.frameHeight * Frame.tileHeight
	
	xmouse = math.min (xmouse, thresh_x2)
	xmouse = math.max (xmouse, thresh_x1)
	ymouse = math.min (ymouse, thresh_y2)
	ymouse = math.max (ymouse, thresh_y1)
	
	local w = xmouse - xsrc
	local h = ymouse - ysrc
	
	if (w < 0) then
		w = math.abs (w)
		xsrc = xsrc - w
	elseif (w == 0) then
		w = 1
	end
	
	if (h < 0) then
		h = math.abs (h)
		ysrc = ysrc - h
	elseif (h == 0) then
		h = 1
	end
	
	--DoCall ("MF_EditorSelect", xsrc, ysrc, (xmouse - xsrc), (ymouse - ysrc) )
	MF_EditorSelect (xsrc, ysrc, w, h)
end

function EditorExpandSnappingSelect (xsrc, ysrc, xmouse, ymouse)
	xsrc = tonumber (xsrc)
	ysrc = tonumber (ysrc)
	xmouse = tonumber (xmouse)
	ymouse = tonumber (ymouse)
	
	if (xmouse < Editor.select_box.orig_x) then
		xsrc = Editor.select_box.orig_x
	end
	if (ymouse < Editor.select_box.orig_y) then
		ysrc = Editor.select_box.orig_y
	end
	
	local tile_sx, tile_sy
	tile_sx, tile_sy = CoordToTile (xsrc, ysrc, 1)
	tile_sx = math.max (tile_sx, 0)
	tile_sy = math.max (tile_sy, 0)
	
	local tile_dx, tile_dy, dummy
	tile_dx, dummy = CoordToTile (xmouse, ysrc, 1)
	dummy, tile_dy = CoordToTile (xsrc, ymouse, 1)

	if (tile_dx == -1 and xmouse > xsrc) then
		tile_dx = Frame.frameWidth
	end
	if (tile_dy == -1 and ymouse > ysrc) then
		tile_dy = Frame.frameHeight
	end
	
	if ( (xmouse - xsrc) % Frame.tileWidth > (Frame.tileWidth / 2) ) then
		tile_dx = math.min (tile_dx + 1, Frame.frameWidth)
	end
	if ( (ymouse - ysrc) % Frame.tileHeight > (Frame.tileHeight / 2) ) then
		tile_dy = math.min (tile_dy + 1, Frame.frameHeight)
	end
	
	tile_dx = math.max (tile_dx, 0)
	tile_dy = math.max (tile_dy, 0)
	
	local w = (tile_dx - tile_sx) * Frame.tileWidth
	local h = (tile_dy - tile_sy) * Frame.tileHeight
	
	if (w < 0) then
		w = math.abs (w)
		xsrc = xsrc - w
	elseif (w == 0) then
		w = 1
	end
	
	if (h < 0) then
		h = math.abs (h)
		ysrc = ysrc - h
	elseif (h == 0) then
		h = 1
	end
	
	--DoCall ("MF_EditorSnappingSelect", xsrc, ysrc, w, h)
	MF_EditorSnappingSelect (xsrc, ysrc, w, h)
end

function EditorFloatingSelect (coordx, coordy, w, h)
	coordx = tonumber (coordx)
	coordy = tonumber (coordy)
	w = tonumber (w)
	h = tonumber (h)
	
	Editor.winstate.select_valid = 0
	
	if (w < 0) then
		w = w * -1
		coordx = coordx - w
	end
	if (h < 0) then
		h = h * -1
		coordy = coordy - h
	end
	
	if (w < Frame.tileWidth and (1 - w) < Frame.tileWidth) then
		--DoCall ("MF_EditorClearSelect")
		MF_EditorClearSelect ()
		return
	end
	if (h < Frame.tileHeight and (1 - h) < Frame.tileHeight) then
		--DoCall ("MF_EditorClearSelect")
		MF_EditorClearSelect ()
		return
	end
	
	Editor.winstate.select_valid = 1

	--DoCall ("MF_EditorFloatingSelect", coordx, coordy, w, h)
	MF_EditorFloatingSelect (coordx, coordy, w, h)
end

function EditorFloatingMousePos (coordx, coordy, mousex, mousey)
	coordx = tonumber (coordx)
	coordy = tonumber (coordy)
	mousex = tonumber (mousex)
	mousey = tonumber (mousey)
	
	Editor.floating.orig_x = coordx
	Editor.floating.orig_y = coordy
	Editor.floating.x = coordx
	Editor.floating.y = coordy
	Editor.floating.xdiff = mousex
	Editor.floating.ydiff = mousey
end

function EditorFloatingSnapping (coordx, coordy, w, h)
	coordx = tonumber (coordx)
	coordy = tonumber (coordy)

	local new_x = Editor.floating.orig_x + (coordx - Editor.floating.xdiff)
	local new_y = Editor.floating.orig_y + (coordy - Editor.floating.ydiff)
	
	Editor.floating.x = new_x
	Editor.floating.y = new_y
	
	local tile_x, tile_y
	tile_x, tile_y = CoordToTile (new_x, new_y, 1)
	tile_x = math.max (tile_x, 0)
	tile_y = math.max (tile_y, 0)
	
	local tile_overx, tile_overy, dummy
	tile_overx, dummy = CoordToTile (new_x + w, new_y, 1)
	dummy, tile_overy = CoordToTile (new_x, new_y + h, 1)
	
	if ( tile_overx == -1 and (tile_x > 0 or (tile_x == 0 and new_x > Editor.floating.orig_x) ) ) then
		tile_x = Frame.frameWidth - (w / Frame.tileWidth)
	end
	if ( tile_overy == -1 and (tile_y > 0 or (tile_y == 0 and new_y > Editor.floating.orig_y) ) ) then
		tile_y = Frame.frameHeight - (h / Frame.tileHeight)
	end

	new_x, new_y = TileToCoord (tile_x, tile_y)
	
	EditorFloatingSelect (new_x, new_y, w, h)
end

function EditorTestAutoScroll ()
	local lx_thresh = Editor.winstate.xmouse - (Editor.winstate.frame_x + Editor.ui.sb_width) 
	local rx_thresh = Editor.winstate.frame_x + Editor.winstate.win_width - Editor.ui.scrollbar_width - Editor.winstate.xmouse
	local ty_thresh = Editor.winstate.ymouse - (Editor.winstate.frame_y + Editor.ui.toolbar_width)
	local by_thresh = Editor.winstate.frame_y + Editor.winstate.win_height - Editor.ui.scrollbar_width - Editor.winstate.ymouse
	
	if (Editor.winstate.select_valid == 0) then
		return
	end
	
	if (lx_thresh <= Editor.ui.scroll_thresh) then
		--DoCall ("MF_EditorSetHScroll", Editor.winstate.hscroll - Editor.ui.tile_size)
		MF_EditorSetHScroll (Editor.winstate.hscroll - Frame.tileWidth)
	elseif (rx_thresh <= Editor.ui.scroll_thresh) then
		--DoCall ("MF_EditorSetHScroll", Editor.winstate.hscroll + Editor.ui.tile_size)
		MF_EditorSetHScroll (Editor.winstate.hscroll + Frame.tileWidth)
	end
	
	if (ty_thresh <= Editor.ui.scroll_thresh) then
		--DoCall ("MF_EditorSetVScroll", Editor.winstate.vscroll - Editor.ui.tile_size)
		MF_EditorSetVScroll (Editor.winstate.vscroll - Frame.tileHeight)
	elseif (by_thresh <= Editor.ui.scroll_thresh) then
		--DoCall ("MF_EditorSetVScroll", Editor.winstate.vscroll + Editor.ui.tile_size)
		MF_EditorSetVScroll (Editor.winstate.vscroll + Frame.tileHeight)
	end	
end


-- Frame Interface ------------------------------------------------------------

function FrameGetProperties ()
	return Frame:GetProperties()
end

-- Profile (Object Class) Interface -------------------------------------------

function ProfileObjectGetAttribs (objid)
	return Active.class:getAttribs (objid)
end

function ProfileObjectUpdateAttribs (objid, name, file, ident)
	Active.class:updateAttribs (objid, name, file, ident)
	SetModified()
end

function ProfileNewObject (name, file, ident)
	Active.class:newObject (name, file, ident)
	SetModified()
end

function ProfileDeleteObject (objid)
	ActiveObjectDelete (objid)
	Active.class:deleteObject (objid)
	SetModified()
end

function ProfileNumObjects ()
	return Active.class.numActives
end

-- Active (Object Instance) Interface -----------------------------------------

function ActiveObjectDelete (objid)
	objid = tonumber (objid)
	for k,v in pairs (Active.inst) do
		if (v.objid == objid) then
			ActiveInstanceDelete (v.instid)
		end
	end
end

function ActiveObjectRedraw (objid)
	objid = tonumber (objid)
	for k,v in pairs (Active.inst) do
		if (v.objid == objid) then
			ActiveInstanceRedraw (v.instid)
		end
	end
end

function ActiveInstanceDelete (instid)
	instid = tonumber (instid)
	Active:clear (instid)
	SetModified()
end

function ActiveInstanceRedraw (instid)
	instid = tonumber (instid)
	local ins = Active.inst[instid]
	if (ins ~= nil) then
		local obj = Active.class.actives[ins.objid]
		Active:_clear (instid)
		Active:_draw (ins.instid, ins.objid, obj.file, ins.x, ins.y, obj.name)
	end
end

function AddActive (objid, dx, dy)
	dx, dy = CoordSubOffset (dx, dy)
	if (dx < 0 or dy < 0) then
		return
	end
	--Active:add (objid, dx, dy)
	
	local alist = { }
	table.insert (alist, { objid = objid, dx = dx, dy = dy } )
	Active:addList (alist)
	
	SetModified()
end

function ActiveDrawIndex ()
	local ix = 0
	sactives = util.paramsort( Active.class.actives, 'name' )
	for k,v in pairs(sactives) do
		--DoCall( "ActiveDrawIndex", v.id, v.name, v.file, (ix % 2), math.floor(ix / 2) )
		MF_ActiveDrawIndex (v.id, v.name, v.file, (ix % 2), math.floor(ix / 2))
		ix = ix + 1
	end
	--DoCall( "ActiveEndIndex", ix )
end

function ActiveStampCoords (xpos, ypos, xsnap, ysnap, width, height)
	xpos = tonumber (xpos)
	ypos = tonumber (ypos)
	width = tonumber (width)
	height = tonumber (height)
	xsnap = tonumber (xsnap)
	ysnap = tonumber (ysnap)
	
	local x_tile, y_tile = CoordToTile (xpos, ypos)
	if (x_tile == -1 or y_tile == -1) then
		return xpos, ypos
	end
	
	local xadj, yadj = TileToCoord (x_tile, y_tile)
	
	local xdiff = xpos - xadj
	local ydiff = ypos - yadj
	
	if (xsnap == 1) then
		xpos = xadj
	elseif (xsnap == 2) then
		xpos = xadj + (Frame.tileWidth - (width % Frame.tileWidth) )
		if ( (width % Frame.tileWidth) == 0) then
			xpos = xpos - Frame.tileWidth
		end
	end
	
	if (ysnap == 1) then
		ypos = yadj
	elseif (ysnap == 2) then
		ypos = yadj + (Frame.tileHeight - (height % Frame.tileHeight) )
		if ( (height % Frame.tileHeight) == 0) then
			ypos = ypos - Frame.tileHeight
		end
	end
	
	return xpos, ypos
end

function ActiveSelectObjects (x, y, w, h)
	ActiveClearSelection ()
	
	x, y = CoordSubOffset (x, y)
	
	for k,v in pairs (Active.inst) do
		if (v.x >= x and v.x < x + w and v.y >= y and v.y < y + h) then
			ActiveAddSingleSelect (v.instid)
		end
	end
end

function ActiveSingleSelect (instid)
	ActiveClearSelection ()
	ActiveAddSingleSelect (instid)
end

function ActiveAddSingleSelect (instid)
	if (Active.inst[instid] ~= nil) then
		Editor.selected_objects[instid] = instid
		MF_ActiveSelectObject (instid)
	end
end

function ActiveSelectAll (objid)
	ActiveClearSelection ()
	for k,v in pairs (Active.inst) do
		if (v.objid == objid) then
			ActiveAddSingleSelect (v.instid)
		end
	end
end

function ActiveClearSelection ()
	for k,v in pairs (Editor.selected_objects) do
		MF_ActiveUnselectObject (v)
	end
	
	Editor.selected_objects = {}
end

function ActiveDeleteSelection ()
	for k,v in pairs (Editor.selected_objects) do
		MF_ActiveUnselectObject (v)
		MF_ActiveDeleteObject (v)
		--ActiveInstanceDelete (v)
	end
	
	Active:clearList (Editor.selected_objects)
	
	Editor.selected_objects = {}
	SetModified ()
end

function ActiveStartDragSelection (xmouse, ymouse)
	Editor.selected_drag.orig_x = xmouse
	Editor.selected_drag.orig_y = ymouse
end

function ActiveDragSelection (xmouse, ymouse)
	local step_xdiff = xmouse - Editor.selected_drag.xdiff
	local step_ydiff = ymouse - Editor.selected_drag.ydiff
	
	Editor.selected_drag.xdiff = xmouse - Editor.selected_drag.orig_x
	Editor.selected_drag.ydiff = ymouse - Editor.selected_drag.orig_y
	
	for k,v in pairs (Editor.selected_objects) do
		local lx = Active.inst[v].x + Editor.selected_drag.xdiff
		local ly = Active.inst[v].y + Editor.selected_drag.ydiff
		
		lx, ly = CoordAddOffset (lx, ly)
		MF_ActiveSetPosition (v, lx, ly)
	end
	
	--[[Active:moveList (Editor.selected_drag.xdiff, Editor.selected_drag.ydiff, Editor.selected_objects)
	for k,v in pairs (Editor.selected_objects) do
		local lx, ly = CoordAddOffset (Active.inst[v].x, Active.inst[v].y)
		MF_ActiveSetPosition (v, lx, ly)
	end]]
	
	if (Editor.selected_drag.xdiff ~= 0 or Editor.selected_drag.ydiff ~= 0) then
		EditorTestAutoScroll ()
	end
end

function ActiveEndDragSelection ()
	--[[for k,v in pairs (Editor.selected_objects) do
		local lx = Active.inst[v].x + Editor.selected_drag.xdiff
		local ly = Active.inst[v].y + Editor.selected_drag.ydiff
		Active:setPosition (v, lx, ly)
		
		lx, ly = CoordAddOffset (lx, ly)
		MF_ActiveSetPosition (v, lx, ly)
	end]]
	
	Active:moveList (Editor.selected_drag.xdiff, Editor.selected_drag.ydiff, Editor.selected_objects)
	for k,v in pairs (Editor.selected_objects) do
		local lx, ly = CoordAddOffset (Active.inst[v].x, Active.inst[v].y)
		MF_ActiveSetPosition (v, lx, ly)
	end
	
	SetModified ()
end

function ActiveGetInstValueString (instid)
	return Active:getValueString (instid)
end

function ActiveSetInstValueString (instid, str)
	local stat, ret = pcall (function() return Active:setValueString (instid, str) end)
	if (stat == false) then
		MF_Error (ret)
	end
	SetModified()
end

function ActiveGetInstAttribs (instid)
	return Active:getAttribs (instid)
end

function ActiveGetObjValueString (objid)
	return Active.class:getValueString (objid)
end

function ActiveSetObjValueString (objid, str)
	local stat, ret = pcall (function() return Active.class:setValueString (objid, str) end)
	if (stat == false) then
		MF_Error (ret)
	end
	SetModified()
end


-- Tile Interface -------------------------------------------------------------

function TilePaneSnappingSelect (xmouse, ymouse)
	if (xmouse <= Editor.winstate.frame_x or
		xmouse >= Editor.winstate.frame_x + Editor.ui.sb_width - Editor.ui.scrollbar_width - 2 or
		ymouse <= Editor.winstate.frame_y + Editor.ui.toolbar_width + 1 or
		ymouse >= Editor.winstate.frame_y + Editor.winstate.win_height - 1) then
		return
	end
	
	local tsx = Frame.tileWidth + 1
	local tsy = Frame.tileHeight + 1
	local tile_x = math.floor ( (xmouse - Editor.winstate.frame_x - ( (xmouse - Editor.winstate.frame_x) % tsx) ) / tsx)
	
	local diff = ymouse - Editor.winstate.frame_y - Editor.ui.toolbar_width - 1 + Editor.winstate.sscroll
	local tile_y = math.floor ( (diff - (diff % tsy) ) / tsy)
	
	local coord_x = Editor.winstate.frame_x + (Frame.tileWidth + 1) * tile_x
	local coord_y = Editor.winstate.frame_y + Editor.ui.toolbar_width + 1 - Editor.winstate.sscroll + (Frame.tileHeight + 1) * tile_y
	
	Editor.tile_pane_x = tile_x
	Editor.tile_pane_y = tile_y
	Editor.tile_pane_w = 1
	Editor.tile_pane_h = 1
	
	MF_TilePaneSelect (coord_x, coord_y, Frame.tileWidth+1, Frame.tileHeight+1)
end

function TilePaneExpandSnappingSelect (x, y, xmouse, ymouse)	
	if (xmouse <= Editor.winstate.frame_x or
		xmouse >= Editor.winstate.frame_x + Editor.ui.sb_width - Editor.ui.scrollbar_width - 2 or
		ymouse <= Editor.winstate.frame_y + Editor.ui.toolbar_width + 1 or
		ymouse >= Editor.winstate.frame_y + Editor.winstate.win_height - 1) then
		return
	end
	
	local tsx = Frame.tileWidth + 1
	local tsy = Frame.tileHeight + 1
	local tile_x_orig = math.floor ( (x - Editor.winstate.frame_x - ( (x - Editor.winstate.frame_x) % tsx) ) / tsx)
	local tile_x = math.floor ( (xmouse - Editor.winstate.frame_x - ( (xmouse - Editor.winstate.frame_x) % tsx) ) / tsx)
	
	local diff_orig = ymouse - Editor.winstate.frame_y - Editor.ui.toolbar_width - 1 + Editor.winstate.sscroll
	local diff = y - Editor.winstate.frame_y - Editor.ui.toolbar_width - 1 + Editor.winstate.sscroll
	local tile_y_orig = math.floor ( (diff_orig - (diff_orig % tsy) ) / tsy)
	local tile_y = math.floor ( (diff - (diff % tsy) ) / tsy)
	
	local coord_x = Editor.winstate.frame_x + (Frame.tileWidth + 1) * tile_x_orig
	local coord_y = Editor.winstate.frame_y + Editor.ui.toolbar_width + 1 - Editor.winstate.sscroll + (Frame.tileHeight + 1) * tile_y_orig
	
	Editor.tile_pane_w = math.max (1, math.abs (tile_x - tile_x_orig))
	Editor.tile_pane_h = math.max (1, math.abs (tile_y - tile_y_orig))
	
	MF_TilePaneSelect (x+1, y+1, Editor.tile_pane_w * (Frame.tileWidth+1), Editor.tile_pane_h * (Frame.tileHeight+1) )
end

function TilePaneCoordToTile (xmouse, ymouse)
	xmouse = tonumber (xmouse)
	ymouse = tonumber (ymouse)
	
	if (xmouse <= Editor.winstate.frame_x or
		xmouse >= Editor.winstate.frame_x + Editor.ui.sb_width - Editor.ui.scrollbar_width - 2 or
		ymouse <= Editor.winstate.frame_y + Editor.ui.toolbar_width + 1 or
		ymouse >= Editor.winstate.frame_y + Editor.winstate.win_height - 1) then
		return
	end

	local tsx = Frame.tileWidth + 1
	local tsy = Frame.tileHeight + 1
	local tile_x = math.floor ( (xmouse - Editor.winstate.frame_x - ( (xmouse - Editor.winstate.frame_x) % tsx) ) / tsx)
	
	local diff = ymouse - Editor.winstate.frame_y - Editor.ui.toolbar_width - 1 + Editor.winstate.sscroll
	local tile_y = math.floor ( (diff - (diff % tsy) ) / tsy)
	
	Editor.tile_pane_x = tile_x
	Editor.tile_pane_y = tile_y
end

function TileCreateSlice (mtype, xcoord, ycoord, w, h)
	local x_tile, y_tile
	x_tile, y_tile = CoordToTile (xcoord, ycoord, 1)
	w = tonumber (w) / Frame.tileWidth
	h = tonumber (h) / Frame.tileHeight
	
	local obj = GetObject( mtype )
	if (obj ~= nil) then
		obj.fl_slice = obj:createFloat (x_tile, y_tile, x_tile + w - 1, y_tile + h - 1, 'cut')
	end
	SetModified()
end

function TileDragSlice (mtype, sx, sy, dx, dy)
	local obj = GetObject (mtype)
	if (obj == nil or obj.fl_slice.x == nil) then
		return
	end
	
	sx = tonumber (sx)
	sy = tonumber (sy)
	dx = tonumber (dx)
	dy = tonumber (dy)
	sx, sy = CoordToTile (sx, sy, 1)
	dx, dy = CoordToTile (dx, dy, 1)
	
	local sw = obj.fl_slice.width
	local sh = obj.fl_slice.height
	local diffox = dx - obj.fl_slice.x
	local diffoy = dy - obj.fl_slice.y
	local minx = math.min( sx, dx )
	local miny = math.min( sy, dy )
	local maxx = math.max( sx, dx )
	local maxy = math.max( sy, dy )
	obj.fl_slice.x = obj.fl_slice.x + diffox
	obj.fl_slice.y = obj.fl_slice.y + diffoy
	
	if (dx > sx) then
		RedrawRegion (mtype, sx, miny, dx-1, maxy+sh)
	elseif (dx < sx) then
		RedrawRegion (mtype, dx+sw+1, miny, sx+sw, maxy+sh)
	end
	
	if (dy > sy) then
		RedrawRegion (mtype, minx, sy, maxx+sw, dy-1)
	elseif (dy < sy) then
		RedrawRegion (mtype, minx, dy+sh+1, maxx+sw, sy+sh)
	end
	
	for ix = dx,dx+sw do
		for iy = dy,dy+sh do
			TempClearTile (mtype, ix, iy)
			obj:redrawAt (ix, iy, ix, iy)
			obj:redrawFloatAt (obj.fl_slice, ix-dx, iy-dy, ix, iy)
		end
	end
	SetModified()
end

function TilePasteSlice (mtype)
	local obj = GetObject (mtype)
	if (obj == nil or obj.fl_slice.x == nil) then
		return
	end

	local fl = obj.fl_slice
	obj:pasteFloat (obj.fl_slice)
	RedrawRegion (mtype, fl.x, fl.y, fl.x + fl.width, fl.y + fl.height)
	obj.fl_slice = {}
	SetModified()
end

function TileDeleteSlice (mtype)
	local obj = GetObject (mtype)
	if (obj == nil or obj.fl_slice.x == nil) then
		return
	end
	
	local fl = obj.fl_slice
	RedrawRegion (mtype, fl.x, fl.y, fl.x+fl.width, fl.y+fl.height)
	obj.fl_slice = {}
	SetModified()
end

function TileCut (mtype, x1, y1, w, h)
	w = tonumber (w)
	h = tonumber (h)
	x1, y1 = CoordToTile (x1, y1, 1)
	
	local x2 = x1 + (w / Frame.tileWidth) - 1
	local y2 = y1 + (h / Frame.tileHeight) - 1
	
	local obj = GetObject (mtype)
	if (obj ~= nil) then
		--obj.fl_clipboard = obj:createFloat (x1, y1, x2, y2, 'cut')
		obj.fl_clipboard = obj.fl_slice
		obj.fl_slice = {}
		RedrawRegion (mtype, x1, y1, x2, y2)
		SetModified()
	end
end

function TileCopy (mtype, x1, y1, w, h)
	w = tonumber (w)
	h = tonumber (h)
	x1, y1 = CoordToTile (x1, y1, 1)
	
	local x2 = x1 + (w / Frame.tileWidth) - 1
	local y2 = y1 + (h / Frame.tileHeight) - 1
	
	local obj = GetObject (mtype)
	if (obj ~= nil) then
		--obj.fl_clipboard = obj:createFloat (x1, y1, x2, y2, 'copy')
		obj.fl_clipboard = obj.fl_slice
		TilePasteSlice (mtype)
	end
end

function TilePaste (mtype, x1, y1)
	x1 = tonumber (x1)
	y1 = tonumber (y1)
	
	if (x1 == 0 and y1 == 0) then
		x1 = Editor.winstate.frame_x + Editor.ui.sb_width + Frame.tileWidth * 4
		y1 = Editor.winstate.frame_y + Editor.ui.toolbar_width + Frame.tileHeight * 4
	end
	
	local tile_x, tile_y
	tile_x, tile_y = CoordToTile (x1, y1, 1)
	tile_x = math.max (0, tile_x)
	tile_y = math.max (0, tile_y)
	
	x1, y1 = TileToCoord (tile_x, tile_y)
	
	local obj = GetObject (mtype)
	if (obj ~= nil) then
		obj.fl_slice = obj.fl_clipboard
		Editor.floating.orig_x = x1
		Editor.floating.orig_y = y1
		obj.fl_slice.x = tonumber (x1)
		obj.fl_slice.y = tonumber (y1)
		
		--DoCall ("MF_TileNewSlice", x1, y1, (obj.fl_slice.width + 1) * Frame.tileWidth, (obj.fl_slice.height + 1) * Editor.ui.tile_size)
		MF_TileNewSlice (x1, y1, (obj.fl_slice.width + 1) * Frame.tileWidth, (obj.fl_slice.height + 1) * Frame.tileHeight)
		TileDragSlice (mtype, x1, y1, x1, y1)
	end
	SetModified()
end

function TileFill (mtype, x1, y1, w, h, ...)
	local obj = GetObject (mtype)
	
	x1, y1 = CoordToTile (x1, y1, 1)
	w = w / Frame.tileWidth
	h = h / Frame.tileHeight
	
	obj:fillBlock (x1, y1, Editor.tile_pane_x, Editor.tile_pane_y, w, h, Editor.tile_pane_w, Editor.tile_pane_h, ...)
	SetModified()
end

function AddTilePX (mtype, dx, dy, ...)
	local obj = GetObject (mtype)
	if (dx >= Editor.winstate.frame_x + Editor.winstate.win_width - Editor.ui.scrollbar_width or
		dy >= Editor.winstate.frame_y + Editor.winstate.win_height - Editor.ui.scrollbar_width or
		dx <= Editor.winstate.frame_x + Editor.ui.sb_width or
		dy <= Editor.winstate.frame_y + Editor.ui.toolbar_width) then
		return
	end
	
	dx, dy = CoordToTile (dx, dy)
	if (dx > -1 and dy > -1) then
		--obj:add (dx, dy, Editor.tile_pane_x, Editor.tile_pane_y, ...)
		obj:addBlock (dx, dy, Editor.tile_pane_x, Editor.tile_pane_y, Editor.tile_pane_w, Editor.tile_pane_h, ...)
		SetModified ()
	end
end

function ClearTilePX (mtype, dx, dy)
	local obj = GetObject (mtype)
	if (dx >= Editor.winstate.frame_x + Editor.winstate.win_width - Editor.ui.scrollbar_width or
		dy >= Editor.winstate.frame_y + Editor.winstate.win_height - Editor.ui.scrollbar_width or
		dx <= Editor.winstate.frame_x + Editor.ui.sb_width or
		dy <= Editor.winstate.frame_y + Editor.ui.toolbar_width) then
		return
	end
	
	dx, dy = CoordToTile (dx, dy)
	if (dx > -1 and dy > -1) then
		obj:clear (dx, dy)
		SetModified ()
	end
end

-- Parallax Interface ---------------------------------------------------------

function ParallaxAdd (id, name, z_order, bg_type, bg_start, anim_speed, clamp_x, clamp_y, x_offset, y_offset, file, width, height)
	local newid = Parallax:add (id, name, z_order, bg_type, bg_start, anim_speed, clamp_x, clamp_y, x_offset, y_offset)
	Parallax:addFrame (newid, file, width, height)
	SetModified()
	return newid
end

function ParallaxDelete (id)
	Parallax:dekete (id)
	SetModified()
end

function ParallaxUpdate (id, name, z_order, bg_type, bg_start, anim_speed, clamp_x, clamp_y, x_offset, y_offset, file, width, height)
	Parallax:update (id, name, z_order, bg_type, bg_start, anim_speed, clamp_x, clamp_y, x_offset, y_offset)
	--Parallax:updateFrame (id, 1, file, width, height)
	SetModified()
end

function ParallaxUpdateScrolling (id, ...)
	Parallax:updateScrolling (id, ...)
	SetModified()
end

function ParallaxDrawIndex ()
	local bgs = Parallax:zList()
	local pos_id = 0
	for k,v in pairs (bgs) do
		--DoCall ("MF_ParallaxDrawIndex", pos_id, v.id, v.name, v.file, v.width, v.height, v.z_order)
		MF_ParallaxDrawIndex (pos_id, v.id, v.name, v.images[1].file, v.images[1].width, v.images[1].height, v.z_order)
		pos_id = pos_id + 1
	end
end

function ParallaxGetAttribs (id)
	return Parallax:getAttribs (id)
end

function ParallaxGetAnimation (id)
	return Parallax:getAnimation (id)
end

function ParallaxUpdateAnimation (id, anim_speed, width, height, filestring)
	Parallax:updateAnimation (id, anim_speed, width, height, filestring)
	SetModified()
end

function ParallaxDelete (id)
	Parallax:delete (id)
	SetModified()
end

function ParallaxReOrder (id, dir)
	stat, err = pcall (function() return Parallax:reorder (id, dir) end)
	if (stat == false) then
		MF_Error (err)
	end
	SetModified()
end

-- StringTable Interface ------------------------------------------------------

function StringTableGetAll ()
	StringTable:pushAll()
end

function StringTableClear ()
	StringTable:clear()
	SetModified()
end

function StringTableAdd (id, str)
	StringTable:add (id, str)
	SetModified()
end

-- Zone Interface -------------------------------------------------------------

--[[function ZoneDrawIndex ()
	for i = 1,#Zone.zones do
		local lzone = Zone.zones[i]
		DoCall( "MF_ZoneDrawIndex", i, lzone.name, lzone.x, lzone.y, lzone.width, lzone.height )
	end
end]]

-- Compound Operations --------------------------------------------------------

function ClipRegion( x1, y1, x2, y2 )
	for ix = x1,x2 do
		for iy = y1,y2 do
			Collision:clear( ix, iy )
			Background:clear( ix, iy )
			Foreground:clear( ix, iy )
		end
	end
	
	SetModified()
end

function SelectiveClipRegion( mtype, x1, y1, x2, y2 )
	for ix = x1,x2 do
		for iy = y1,y2 do
			ClearTile( mtype, ix, iy )
		end
	end
	
	SetModified()
end

function FillRegion( mtype, x1, y1, x2, y2, ... )
	for ix = x1,x2 do
		for iy = y1,y2 do
			AddTile( mtype, ix, iy, ... )
		end
	end
	
	SetModified()
end

function TempClearRegion( mtype, x1, y1, x2, y2 )
	for ix = x1,x2 do
		for iy = y1,y2 do
			TempClearTile( mtype, ix, iy )
		end
	end
	return 1
end

function RedrawRegion( mtype, x1, y1, x2, y2 )
	for ix = x1,x2 do
		for iy = y1,y2 do
			TempClearTile( mtype, ix, iy )
			DrawTileAt( mtype, ix, iy, ix, iy )
		end
	end
end

--[[function RedrawZone( zid )
	if ( zid == nil ) then
		zid = Frame.currentZone
	end
	
	zid = tonumber(zid)
	if ( zid < 1 ) then
		zid = 1
	elseif ( zid > #Zone.zones ) then
		zid = #Zone.zones
	end
	local zn = Zone.zones[zid]
	
	for ix = zn.x,zn.x+zn.width-1 do
		for iy = zn.y,zn.y+zn.height-1 do
			Collision:redrawAt( ix, iy, ix, iy )
			Background:redrawAt( ix, iy, ix, iy )
			Foreground:redrawAt( ix, iy, ix, iy )
		end
	end
end]]

function DrawRegionAt( mtype, x1, y1, x2, y2, nx, ny )
	local diffx = nx - x1
	local diffy = ny - y1
	for ix = x1,x2 do
		for iy = y1,y2 do
			DrawTileAt( mtype, ix, iy, ix+diffx, iy+diffy )
		end
	end
end

function FrameSize( nw, nh )
	nw = tonumber(nw)
	nh = tonumber(nh)
	
	local ow = Frame.frameWidth
	local oh = Frame.frameHeight
	
	if (nw == ow and nh == oh) then
		return
	end
	
	-- Maybe we want to abort
	if (nw < ow or nh < oh) then
		local confirm = INT_ConfirmShrinkFrame ()
		if (tonumber (confirm) == 5) then
			nw = ow
			nh = oh
		end
	end		
	
	Frame.frameWidth = nw
	Frame.frameHeight = nh
	
	if ( nw < ow ) then
		ClipRegion( nw, 0, ow, oh )
	end
	if ( nh < oh ) then
		ClipRegion( 0, nh, ow, oh )
	end
	
	--DoCall( "SetFrameSize", nw, nh )
	MF_SetFrameSize (nw, nh)
	SetModified()
end

function FrameTileSize (tw, th)
	if (tw == Frame.tileWidth and th == Frame.tileHeight) then
		return
	end
	
	TempClearRegion ("Collision", 0, 0, Frame.frameWidth - 1, Frame.frameHeight - 1)
	TempClearRegion ("Background", 0, 0, Frame.frameWidth - 1, Frame.frameHeight - 1)
	TempClearRegion ("Foreground", 0, 0, Frame.frameWidth - 1, Frame.frameHeight - 1)
	
	Frame.tileWidth = tw
	Frame.tileHeight = th
	
	RedrawRegion ("Collision", 0, 0, Frame.frameWidth - 1, Frame.frameHeight - 1)
	RedrawRegion ("Background", 0, 0, Frame.frameWidth - 1, Frame.frameHeight - 1)
	RedrawRegion ("Foreground", 0, 0, Frame.frameWidth - 1, Frame.frameHeight - 1)
	
	EditorDrawGrid (Editor.grid.lastx, Editor.grid.lasty, Editor.grid.lastw, Editor.grid.lasth)
	SetModified()
end

function FrameSetColor (color)
	Frame.background_color = color
	SetModified()
end

function GetLevelProperties ()
	return Frame.levelName, Frame.author, Frame.key
end

function SetLevelProperties (name, author, key)
	Frame.levelName = name
	Frame.author = author
	Frame.key = key
	SetModified()
end

-- Primitive Operations and Callbacks -----------------------------------------

function GetObject( mtype )
	if ( mtype == 'Collision' ) then
		return Collision
	elseif ( mtype == 'Background' ) then
		return Background
	elseif ( mtype == 'Foreground' ) then
		return Foreground
	end
end
function AddTile( mtype, ... )
	local obj = GetObject( mtype )
	obj:add( ... )
	SetModified()
end
function ClearTile( mtype, ... )
	local obj = GetObject( mtype )
	if ( obj ~= nil ) then
		obj:clear( ... )
	end
	SetModified()
end
function TempClearTile( mtype, dx, dy )
	local obj = GetObject( mtype )
	obj:_clear( dx, dy )
end
function DrawTileAt( mtype, ... )
	local obj = GetObject( mtype )
	obj:redrawAt( ... )
	SetModified()
end
--[[function AddZone( ... )
	Zone:add( ... )
	EditorRedrawPane()
	SetModified()
end
function ClearZone( ... )
	Zone:clear( ... )
	SetModified()
end
function UpdateZone( ... )
	Zone:update( ... )
	SetModified()
end
--function DrawZoneIndex ()
--	Zone:drawIndex()
--end
function DrawZoneBoxes ()
	Zone:drawBoxes()
end
function RegisterZone( zn )
	Zone:registerZone(zn)
end]]
function ProfileLoad( ... )
	Active.class:loadProfile( ... )
end
function ProfileAddActive( ... )
	Active.class:addActive( ... )
end

function ClearActive( ... )
	for k,v in ipairs{...} do
		Active:clear( v )
	end
	SetModified()
end
function GetActiveAttribs( ... )
	return Active:getAttribs( ... )
end
function SetActiveAttribs( ... )
	Active:setAttribs( ... )
	SetModified()
end
function SetActivePosition( ... )
	Active:setPosition( ... )
	SetModified()
end
function SetActiveChipset( mtype )
	if ( mtype == 'Collision' ) then
		--DoCall( "ActiveChipset", Frame.collisionChip )
	elseif ( mtype == 'Background' ) then
		--DoCall( "ActiveChipset", Frame.backgroundChip )
	elseif ( mtype == 'Foreground' ) then
		--DoCall( "ActiveChipset", Frame.foregroundChip )
	end
end
function GetAllChipsets ()
	--DoCall( "SetChipsets", Frame.collisionChip, Frame.backgroundChip, Frame.foregroundChip )
end
function UpdateChipset( mtype, chip )
	local rt = ""
	if ( mtype == 'Collision' ) then
		Frame.collisionChip = util.basename(chip)
		rt = Frame.collisionChip
	elseif ( mtype == 'Background' ) then
		Frame.backgroundChip = util.basename(chip)
		rt = Frame.backgroundChip
	elseif ( mtype == 'Foreground' ) then
		Frame.foregroundChip = util.basename(chip)
		rt = Frame.foregroundChip
	end
	RedrawRegion (mtype, 0, 0, Frame.frameWidth, Frame.frameHeight)
	return Frame.dataDir .. rt
end
function UpdateChipsetSize (mtype, w, h)
	w = tonumber (w)
	h = tonumber (h)
	if (mtype == 'Collision') then
		Collision.chip_width = w
		Collision.chip_height = h
	elseif (mtype == 'Background') then
		Background.chip_width = w
		Background.chip_height = h
	elseif (mtype == 'Foreground') then
		Foreground.chip_width = w
		Foreground.chip_height = h
	end
end
