-- Mapedit --------------------------------------------------------------------
-- General-purpose map/level editor targeted for MMF2.
-- Copyright 2007 Justin Aquadro
-- ----------------------------------------------------------------------------
-- Loader interface for end-user games and applications
-- ----------------------------------------------------------------------------

local util = require "./lua/util"
local mio = require "./lua/mio"
local tile = require "./lua/tile"
local frame = require "./lua/frame"
local zone = require "./lua/zone"
local active = require "./lua/active"
local parallax = require "./lua/parallax"

Frame = frame.FrameClass()
Zone = zone.ZoneClass()
Collision = tile.CollisionClass()
Background = tile.BackgroundClass()
Foreground = tile.ForegroundClass()
Active = active.ActiveClass()
Parallax = parallax.ParallaxClass()

Collision.frame = Frame
Background.frame = Frame
Foreground.frame = Frame

IOC = mio.IOClass()
IOC:registerClass( Frame ) -- This MUST come first
--IOC:registerClass( Zone ) -- This MUST come second
IOC:registerClass( Collision )
IOC:registerClass( Background )
IOC:registerClass( Foreground )
IOC:registerClass( Active )
IOC:registerClass( Parallax )

-- Import / Export ------------------------------------------------------------

function Import( fname, key )
	local stat, err = pcall (function() return IOC:import( fname, key ) end)
	if (stat == false) then
		MF_Error (err)
	end
end

last_x = 0
last_y = 0
first_render = 0

-- Remaining Public Interface -------------------------------------------------

-- Draw an entire tile plane to the frame
function FrameRender (mtype)
	RedrawRegion (mtype, 0, 0, Frame.frameWidth, Frame.frameHeight)
end

-- Get a variable number of values from an object instance.
function GetInstValue (instid, ...)
	local vlist = {}
	for k,v in ipairs{...} do
		table.insert (vlist, Active.getValue (instid, v) )
	end
	
	return unpack (vlist)
end

-- Get all values for an object instance in serialized form
function GetInstValueString (instid)
	return Active:getValueString (instid)
end

-- Get a variable number of values from an object (class).
function GetObjValue (objid, ...)
	local vlist = {}
	for k,v in ipairs{...} do
		table.insert (vlist, Active.class:getValue (objid, v) )
	end
	
	return unpack (vlist)
end

-- Get all object (class) values for an object in serialized form
function GetObjValueString (objid)
	return Active.class:getValueString (objid)
end

-- Dynamically draw the visible part of tile plane(s) to the frame, erasing hidden areas.
function LocalRender (x, y, w, h)
	x = tonumber(x)
	y = tonumber(y)
	w = tonumber(w)
	h = tonumber(h)
	
	local base_x = math.floor (x / 16)
	local base_y = math.floor (y / 16)
	local base_w = math.ceil (w / 16)
	local base_h = math.ceil (h / 16)
	
	local diff_x = base_x - last_x
	local diff_y = base_y - last_y

	if (diff_x == 0 and diff_y == 0 and first_render == 1) then
		return
	end
	
	local draw_distance = 2
	
	local diff_px = math.abs (diff_x)
	local diff_py = math.abs (diff_y)
	local full_w = base_w + (draw_distance * 2)
	local full_h = base_h + (draw_distance * 2)
	
	if (first_render == 0) then
		diff_px = base_w
		diff_py = base_h
		first_render = 1
	end
	
	-- Special case, when the regions do not overlap
	if (diff_px >= base_w or diff_py >= base_h) then
		local clear_sx = last_x - draw_distance
		local clear_sy = last_y - draw_distance
		--TempClearRegion ('Collision', clear_sx, clear_sy, clear_sx + full_w, clear_sy + full_h)
		TempClearRegion ('Background', clear_sx, clear_sy, clear_sx + full_w, clear_sy + full_h)
		TempClearRegion ('Foreground', clear_sx, clear_sy, clear_sx + full_w, clear_sy + full_h)
		
		local draw_sx = base_x - draw_distance
		local draw_sy = base_y - draw_distance
		--RedrawRegion ('Collision', draw_sx, draw_sy, draw_sx + full_w, draw_sy + full_h)
		RedrawRegion ('Background', draw_sx, draw_sy, draw_sx + full_w, draw_sy + full_h)
		RedrawRegion ('Foreground', draw_sx, draw_sy, draw_sx + full_w, draw_sy + full_h)
		
		last_x = base_x
		lasy_y = base_y
		
		return
	end	
	
	-- Un-render tiles
	local clear_sx = last_x - draw_distance
	if (diff_x < 0) then
		clear_sx = base_x + base_w + draw_distance
	end
	
	local clear_sy = last_y - draw_distance
	if (diff_y < 0) then
		clear_sy = base_y + base_h + draw_distance
	end
	
	if (diff_x ~= 0) then
		local hend = clear_sx + math.abs (diff_x) - 1
		local vstart = last_y - draw_distance
		local vend = vstart + base_h + (draw_distance * 2) - 1
		--TempClearRegion ('Collision', clear_sx, vstart, hend, vend)
		TempClearRegion ('Background', clear_sx, vstart, hend, vend)
		TempClearRegion ('Foreground', clear_sx, vstart, hend, vend)
	end
	
	if (diff_y ~= 0) then
		local hstart = last_x - draw_distance
		local hend = hstart + base_w + (draw_distance * 2) - 1
		local vend = clear_sy + math.abs (diff_y) - 1
		--TempClearRegion ('Collision', hstart, clear_sy, hend, vend)
		TempClearRegion ('Background', hstart, clear_sy, hend, vend)
		TempClearRegion ('Foreground', hstart, clear_sy, hend, vend)
	end
	
	-- Render tiles
	local draw_sx = last_x + base_w + draw_distance
	if (diff_x < 0) then
		draw_sx = last_x - draw_distance + diff_x
	end
	
	local draw_sy = last_y + base_h + draw_distance
	if (diff_y < 0) then
		draw_sy = last_y - draw_distance + diff_y
	end
	
	if (diff_x ~= 0) then
		local hend = draw_sx + math.abs (diff_x) - 1
		local vstart = base_y - draw_distance
		local vend = vstart + base_h + (draw_distance * 2) - 1
		--RedrawRegion ('Collision', draw_sx, vstart, hend, vend)
		RedrawRegion ('Background', draw_sx, vstart, hend, vend)
		RedrawRegion ('Foreground', draw_sx, vstart, hend, vend)
	end
	
	if (diff_y ~= 0) then
		local hstart = base_x - draw_distance
		local hend = hstart + base_w + (draw_distance * 2) - 1
		local vend = draw_sy + math.abs (diff_y) - 1
		--RedrawRegion ('Collision', hstart, draw_sy, hend, vend)
		RedrawRegion ('Background', hstart, draw_sy, hend, vend)
		RedrawRegion ('Foreground', hstart, draw_sy, hend, vend)
	end
	
	last_x = base_x
	last_y = base_y	
end

-- Creates and initializes all of the backdrop objects for the level
function LoadBackdrops (frame_x, frame_y, frame_w, frame_h, window_w, window_h)
	local bgs = Parallax:zList()
	for k,v in pairs(bgs) do
		local status, msg = pcall (function() return Parallax:load (k, frame_x, frame_y, frame_w, frame_h, window_w, window_h) end)
		if (status == false) then
			MF_Error (msg)
		end
	end
end

-- Updates all backdrop positions based on current frame, window, and coordinate data
function UpdateBackdrops (frame_x, frame_y, frame_w, frame_h, window_w, window_h)
	local bgs = Parallax:zList()
	for k,v in pairs(bgs) do
		Parallax:move (k, frame_x, frame_y, frame_w, frame_h, window_w, window_h)
	end
end

-- ----------------------------------------------------------------------------

-- Fetches the filenames of the three chipsets for the level
function GetAllChipsets ()
	--DoCall( "CB_GetChipsets", Frame.collisionChip, Frame.backgroundChip, Frame.foregroundChip )
	return Frame.dataDir .. Frame.collisionChip, Frame.dataDir .. Frame.backgroundChip, Frame.dataDir .. Frame.foregroundChip
end






-- Old Interface (Needs cleanup) ----------------------------------------------

--[[function RegisterZone( zn )
	Zone:registerZone(zn)
end]]

-- Compound Operations --------------------------------------------------------

function ClipRegion( x1, y1, x2, y2 )
	for ix = x1,x2 do
		for iy = y1,y2 do
			Collision:clear( ix, iy )
			Background:clear( ix, iy )
			Foreground:clear( ix, iy )
		end
	end
end

function SelectiveClipRegion( mtype, x1, y1, x2, y2 )
	for ix = x1,x2 do
		for iy = y1,y2 do
			ClearTile( mtype, ix, iy )
		end
	end
end

function FillRegion( mtype, x1, y1, x2, y2, ... )
	for ix = x1,x2 do
		for iy = y1,y2 do
			AddTile( mtype, ix, iy, ... )
		end
	end
end

function TempClearRegion( mtype, x1, y1, x2, y2 )
	for ix = x1,x2 do
		for iy = y1,y2 do
			TempClearTile( mtype, ix, iy )
		end
	end
	return 1
end

function RedrawRegion( mtype, x1, y1, x2, y2 )
	for ix = x1,x2 do
		for iy = y1,y2 do
			TempClearTile( mtype, ix, iy )
			DrawTileAt( mtype, ix, iy, ix, iy )
		end
	end
end

--[[function RedrawZone( zid )
	if ( zid == nil ) then
		zid = Frame.currentZone
	end
	
	zid = tonumber(zid)
	if ( zid < 1 ) then
		zid = 1
	elseif ( zid > #Zone.zones ) then
		zid = #Zone.zones
	end
	local zn = Zone.zones[zid]
	
	for ix = zn.x,zn.x+zn.width-1 do
		for iy = zn.y,zn.y+zn.height-1 do
			Collision:redrawAt( ix, iy, ix, iy )
			Background:redrawAt( ix, iy, ix, iy )
			Foreground:redrawAt( ix, iy, ix, iy )
		end
	end
end]]

function DrawRegionAt( mtype, x1, y1, x2, y2, nx, ny )
	local diffx = nx - x1
	local diffy = ny - y1
	for ix = x1,x2 do
		for iy = y1,y2 do
			DrawTileAt( mtype, ix, iy, ix+diffx, iy+diffy )
		end
	end
end

	
function FrameSize( nw, nh )
	nw = tonumber(nw)
	nh = tonumber(nh)
	
	local ow = Frame.frameWidth
	local oh = Frame.frameHeight
	Frame.frameWidth = nw
	Frame.frameHeight = nh
	
	if ( nw < ow ) then
		ClipRegion( nw, 0, ow, oh )
	end
	if ( nh < oh ) then
		ClipRegion( 0, nh, ow, oh )
	end
	
	--DoCall( "SetFrameSize", nw, nh )
end

-- Primitive Operations and Callbacks -----------------------------------------

function GetObject( mtype )
	if ( mtype == 'Collision' ) then
		return Collision
	elseif ( mtype == 'Background' ) then
		return Background
	elseif ( mtype == 'Foreground' ) then
		return Foreground
	end
end
function AddTile( mtype, ... )
	local obj = GetObject( mtype )
	obj:add( ... )
end
function ClearTile( mtype, ... )
	local obj = GetObject( mtype )
	if ( obj ~= nil ) then
		obj:clear( ... )
	end
end
function TempClearTile( mtype, dx, dy )
	local obj = GetObject( mtype )
	obj:_clear( dx, dy )
end
function DrawTileAt( mtype, ... )
	local obj = GetObject( mtype )
	obj:redrawAt( ... )
end
--[[function AddZone( ... )
	Zone:add( ... )
end
function ClearZone( ... )
	Zone:clear( ... )
end
function RegisterZone( zn )
	Zone:registerZone(zn)
end]]
function AddActive( ... )
	Active:add( ... )
end
function ClearActive( ... )
	for k,v in ipairs{...} do
		Active:clear( v )
	end
end
function GetActiveAttribs( ... )
	Active:getAttribs( ... )
end
function SetActiveAttribs( ... )
	Active:setAttribs( ... )
end
function SetActivePosition( ... )
	Active:setPosition( ... )
end
function SetActiveChipset( mtype )
	if ( mtype == 'Collision' ) then
		DoCall( "ActiveChipset", Frame.collisionChip )
	elseif ( mtype == 'Background' ) then
		DoCall( "ActiveChipset", Frame.backgroundChip )
	elseif ( mtype == 'Foreground' ) then
		DoCall( "ActiveChipset", Frame.foregroundChip )
	end
end

function UpdateChipset( mtype, chip )
	if ( mtype == 'Collision' ) then
		Frame.collisionChip = util.basename(chip)
	elseif ( mtype == 'Background' ) then
		Frame.backgroundChip = util.basename(chip)
	elseif ( mtype == 'Foreground' ) then
		Frame.foregroundChip = util.basename(chip)
	end
end
