local util = require "./lua/util"
local protect = require "./lua/protect"

local io = io
local table = table
local string = string

local setmetatable = setmetatable
local pairs = pairs
local pcall = pcall
local DoCall = DoCall

local MF_Error = MF_Error
local MF_MissingFile = MF_MissingFile
local MF_InvalidFile = MF_InvalidFile
local INT_Password = INT_Password

module (...)

-- ----------------------------------------------------------------------------

function IOClass ()
	return Mio:new()
end

Mio = {
	filep = { mime = "MLEV", emime = "MLXV", ver = 10, hsz = 20, compat = "TLV1" },
	classList = {},
}

function Mio:new ()
	obj = {}
	setmetatable (obj, self)
	self.__index = self
	return obj
end

function Mio:registerClass (clo)
	table.insert( self.classList, clo )
end
	
function Mio:import (fname, key)
	local handle = io.open( fname, "rb" )
	if (handle == nil) then
		MF_MissingFile (fname)
		return
	end
	
	io.input( handle )
	
	local data_file = io.read("*all")
	local data_idx = 0

	local blockDesc = string.sub (data_file, data_idx+1, data_idx+8) --io.read(8)
	data_idx = data_idx + 8
	
	local mime = string.sub(blockDesc, 1, 4)
	if (mime == self.filep.emime) then
		if (key == nil) then
			key = INT_Password ()
		end
		data_file = protect.decrypt (key, string.sub (data_file, 5, string.len(data_file)) )
		blockDesc = string.sub (data_file, 1, 8)
		mime = string.sub(blockDesc, 1, 4)
	end
	
	if ( mime ~= self.filep.mime ) then
		MF_InvalidFile (fname)
		return
	end
	
	local ver = util.unu16(string.sub(blockDesc, 5, 6))
	local hsz = util.unu16(string.sub(blockDesc, 7, 8))

	local globData = string.sub (data_file, data_idx+1, data_idx+hsz-8) --io.read(hsz-8)
	data_idx = data_idx + hsz - 8
	
	local fileSize = util.unu32(string.sub(globData, 1, 4))
	local regBlocks = util.unu16(string.sub(globData, 5, 6))
	local blockDefnSz = util.unu16(string.sub(globData, 7, 8))
	
	if (ver >= 8) then
		local compat = string.sub(globData, 9, 12)
		if (compat ~= self.filep.compat) then
			MF_Error ("Modified TLV detected; cannot load file")
		end
	end
	
	local blockOffsets = string.sub (data_file, data_idx+1, data_idx + regBlocks * blockDefnSz) --io.read(regBlocks * blockDefnSz)
	data_idx = data_idx + regBlocks * blockDefnSz

	for i = 1,regBlocks do
		local base = (i-1) * blockDefnSz
		local ofst = util.unu32(string.sub(blockOffsets, base+1, base+4))
		local sLen = util.unu32(string.sub(blockOffsets, base+5, base+8))
		
		--handle:seek("set", ofst)
		data_idx = ofst
		
		local sBlockDesc = string.sub (data_file, data_idx+1, data_idx+8) --io.read(8)
		data_idx = data_idx + 8
		
		local sMime = string.sub(sBlockDesc, 1, 4)
		local sVer = util.unu16(string.sub(sBlockDesc, 5, 6))
		local sHsz = util.unu16(string.sub(sBlockDesc, 7, 8))
		
		local sBlockData = string.sub (data_file, data_idx+1, data_idx+sLen-8) --io.read(sLen-8)
		data_idx = data_idx + sLen - 8
		
		for k,v in pairs( self.classList ) do
			if ( sMime == v.filep.mime and v.import ~= nil ) then
				v:import( sBlockData, sVer )
			end
		end
	end
	
	io.input( io.stdin )
end
		
function Mio:export (fname, key)
		
	-- Some file parameters - global data
	local mime = self.filep.mime
	local ver = self.filep.ver
	local hsz = self.filep.hsz
	local compat = self.filep.compat
	local szAll = 0
	local regBlocks = #self.classList
	local blockDefnSz = 8
	local globBlock = ""
	local tempBlock = ""
	local blockData = ""
	local globSize = hsz + blockDefnSz*regBlocks

	for k,v in pairs( self.classList ) do
		if ( v.export ~= nil ) then
			tempBlock = v:export()
			--tempBlock = v:export()
			globBlock = globBlock .. util.u32( globSize + #blockData ) .. util.u32( #tempBlock )
			blockData = blockData .. tempBlock
		end
	end
	
	szAll = globSize + #blockData
	globBlock2 = mime .. util.u16(ver) .. util.u16(hsz) .. util.u32(szAll) .. util.u16(regBlocks) .. 
			     util.u16(blockDefnSz) .. compat .. globBlock
	
	local handle = io.open( fname, "wb" )
	io.output( handle )
	
	local enc = ""
	if (key == nil or key == "") then
		enc = globBlock2 .. blockData
	else
		enc = "MLXV" .. protect.encrypt (key, globBlock2 .. blockData)
	end
	
	--io.write( globBlock2 .. blockData )
	io.write (enc)
	io.flush()
	io.close()
	io.output( io.stdout )
end
	
function Mio:reset ()
	for k,v in pairs( self.classList ) do
		if ( v.reset ~= nil ) then
			v:reset()
		end
	end
end
	
function Mio:dbg( fname )
	local handle = io.open( fname .. ".dbg", "wb" )
	io.output( handle )
	
	for k,v in pairs( self.classList ) do
		io.write( v.filep.mime .. " => " .. util.tostring(v) )
	end
	
	io.flush()
	io.close()
	io.output( io.stdout )
end