local util = require "./lua/util"

local math = math
local table = table
local string = string

local setmetatable = setmetatable
local ipairs = ipairs
local pairs = pairs
local DoCall = DoCall

local MF_ClearParallax = MF_ClearParallax
local MF_PasteParallax = MF_PasteParallax
local MF_LoadBackdrop = MF_LoadBackdrop
local MF_AddBackdropFrame = MF_AddBackdropFrame
local MF_MoveBackdrop = MF_MoveBackdrop
local MF_SwitchBackdropFrame = MF_SwitchBackdropFrame
local MF_Error = MF_Error
local MF_Msg = MF_Msg

module (...)

-- ----------------------------------------------------------------------------

function ParallaxClass ()
	return Parallax:new()
end

Parallax = {
	filep = { mime = "PARA", ver = 1, szBlock = 38, szPara = 96, szImage = 68, szFrame = 4 },
	backdrops = {},
	lastIndex = 0,
	numBackdrops = 0,
}

function Parallax:new ()
	obj = {}
	setmetatable (obj, self)
	self.__index = self
	return obj
end

function Parallax:add (id, name, z_order, bg_type, bg_start, anim_speed, clamp_x, clamp_y, x_offset, y_offset)
	id = util.tonumber (id)
	z_order = util.tonumber (z_order)
	
	if (id == 0) then
		id = self.lastIndex + 1
	end
	
	if (id > self.lastIndex) then
		self.lastIndex = id
	end
	
	if (z_order == 0) then
		z_order = id
	end
	
	local newpara = {
		id = id,
		name = name,
		x_repeat = 0,
		y_repeat = 0,
		x_scroll = 0,
		y_scroll = 0,
		x_auto = 0,
		y_auto = 0,
		x_coef = 0,
		y_coef = 0,
		x_speed = 0,
		y_speed = 0,
		auto_scroll = 0,
		cur_frame = 1,
		z_order = z_order,
		bg_start = util.tonumber (bg_start),
		bg_type = util.tonumber (bg_type),
		anim_speed = util.tonumber (anim_speed),
		clamp_x = util.tonumber (clamp_x),
		clamp_y = util.tonumber (clamp_y),
		x_offset = util.tonumber (x_offset),
		y_offset = util.tonumber (y_offset),
		images = { },
		frames = { }
	}

	self.backdrops[id] = newpara
	self.numBackdrops = self.numBackdrops + 1
	
	return self.lastIndex
end

function Parallax:delete (id)
	id = util.tonumber (id)
	if (self.backdrops[id] == nil) then
		return
	end
	
	self.backdrops[id] = nil
	self.numBackdrops = self.numBackdrops - 1
end

function Parallax:addFrame (id, file, width, height)
	id = util.tonumber (id)
	if (self.backdrops[id] == nil) then
		MF_Error ("Parallax:addFrame: Attempt to read nil object at index " .. id)
		return
	end
	
	local frame_index = #self.backdrops[id].frames + 1
	local file_index = 0
	for i = 1, #self.backdrops[id].images do
		if (self.backdrops[id].images[i].file == file) then
			self.backdrops[id].frames[frame_index] = i
			file_index = i
		end
	end
	
	if (file_index == 0) then
		file_index = #self.backdrops[id].images + 1
		self.backdrops[id].images[file_index] = {
			file = file,
			width = util.tonumber (width),
			height = util.tonumber (height)
		}
		self.backdrops[id].frames[frame_index] = file_index
	end
end

function Parallax:updateImage (id, imgid, file, width, height)
	id = util.tonumber (id)
	if (self.backdrops[id] == nil) then
		MF_Error ("Parallax:updateFrame: Attempt to read nil object at index " .. id)
		return
	end 
	
	imgid = util.tonumber (frameid)
	if (self.backdrops[id].images[imgid] == nil) then
		MF_Error ("Parallax:updateImage: Attempt to read nil image at index " .. imgid)
		return
	end

	self.backdrops[id].images[imgid].file = file
	self.backdrops[id].images[imgid].width = util.tonumber (width)
	self.backdrops[id].images[imgid].height = util.tonumber (height)
end

function Parallax:clearFrames (id)
	id = util.tonumber (id)
	if (self.backdrops[id] == nil) then
		MF_Error ("Parallax:clearFrames: Attempt to read nil object at index " .. id)
		return
	end
	
	self.backdrops[id].images = {}
	self.backdrops[id].frames = {}
end

function Parallax:delete (id)
	id = util.tonumber (id)
	if (self.backdrops[id] ~= nil) then
		self.backdrops[id] = nil
		self.numBackdrops = self.numBackdrops - 1
	end
end

function Parallax:reorder (id, dir)
	id = util.tonumber (id)
	if (self.backdrops[id] == nil) then
		return
	end
	
	local prev = 0
	local next = 0
	local id_found = 0
	local bgs = self:zList()
	for k,v in pairs (bgs) do
		if (id_found == 1 and next == 0) then
			next = v.id
		end
		
		if (v.id == id) then
			id_found = 1
		end
		
		if (id_found == 0) then
			prev = v.id
		end
	end
	
	if (dir == 0 and prev ~= 0) then
		local temp = self.backdrops[id]
		self.backdrops[id] = self.backdrops[prev]
		self.backdrops[prev] = temp
		self.backdrops[id].id = id
		self.backdrops[prev].id = prev
	elseif (dir == 1 and next ~= 0) then
		local temp = self.backdrops[id]
		self.backdrops[id] = self.backdrops[next]
		self.backdrops[next] = temp
		self.backdrops[id].id = id
		self.backdrops[next].id = next
	end
end

function Parallax:getAttribs (id)
	id = util.tonumber(id)
	local i = self.backdrops[id]
	if (i == nil) then
		MF_Error ("Parallax:getAttribs: Attempt to read nil object at index " .. id)
		return
	end
	
	return id, i.name, i.images[1].file, i.images[1].width, i.images[1].height, i.x_scroll, i.x_repeat, i.x_auto, i.x_coef, i.x_speed, i.y_scroll, i.y_repeat, i.y_auto, i.y_coef, i.y_speed, i.z_order, i.bg_type, i.bg_start, i.anim_speed, i.clamp_x, i.clamp_y, i.x_offset, i.y_offset
end

function Parallax:getAnimation (id)
	id = util.tonumber(id)
	local i = self.backdrops[id]
	if (i == nil) then
		MF_Error ("Parallax:getAnimation: Attempt to read nil object at index " .. id)
		return
	end
	
	local filestring = i.images[1].file
	for j = 2, #i.images do
		filestring = filestring .. "|" .. i.images[j].file
	end
	
	return id, i.anim_speed, i.images[1].width, i.images[1].height, filestring
end

function Parallax:updateAnimation (id, anim_speed, width, height, filestring)
	id = util.tonumber(id)
	if (self.backdrops[id] == nil) then
		MF_Error ("Parallax:updateAnimation: Attempt to read nil object at index " .. id)
		return
	end
	
	self:clearFrames (id)

	local files = util.str_split ("|", filestring)
	for i = 1, #files do
		self:addFrame (id, files[i], width, height)
	end
	
	self.backdrops[id].anim_speed = anim_speed	
end

function Parallax:update (id, name, z_order, bg_type, bg_start, anim_speed, clamp_x, clamp_y, x_offset, y_offset)
	id = util.tonumber (id)
	z_order = util.tonumber (z_order)
	anim_speed = util.tonumber (anim_speed)
	
	if (self.backdrops[id] == nil) then
		MF_Error ("Parallax:update: Attempt to update a nil object at index " .. id)
		return
	end

	self.backdrops[id].name = name
	self.backdrops[id].bg_type = util.tonumber (bg_type)
	self.backdrops[id].bg_start = util.tonumber (bg_start)
	self.backdrops[id].clamp_x = util.tonumber (clamp_x)
	self.backdrops[id].clamp_y = util.tonumber (clamp_y)
	self.backdrops[id].x_offset = util.tonumber (x_offset)
	self.backdrops[id].y_offset = util.tonumber (y_offset)

	if (anim_speed > 0) then
		self.backdrops[id].anim_speed = anim_speed
	end
	
	if (z_order ~= 0) then
		self.backdrops[id].z_order = z_order
	end
end

function Parallax:updateScrolling (id, x_scroll, x_repeat, x_auto, x_coef, x_speed, y_scroll, y_repeat, y_auto, y_coef, y_speed)
	id = util.tonumber (id)

	if (self.backdrops[id] == nil) then
		MF_Error ("Parallax:updateScrolling: Attempt to update a nil object at index " .. id)
		return
	end
	
	self.backdrops[id].x_coef = util.tonumber (x_coef)
	self.backdrops[id].y_coef = util.tonumber (y_coef)
	self.backdrops[id].x_speed = util.tonumber (x_speed)
	self.backdrops[id].y_speed = util.tonumber (y_speed)
	self.backdrops[id].x_scroll = util.tonumber (x_scroll)
	self.backdrops[id].y_scroll = util.tonumber (y_scroll)
	self.backdrops[id].x_repeat = util.tonumber (x_repeat)
	self.backdrops[id].y_repeat = util.tonumber (y_repeat)
	self.backdrops[id].x_auto = util.tonumber (x_auto)
	self.backdrops[id].y_auto = util.tonumber (y_auto)
end

function Parallax:load (id, frame_x, frame_y, frame_w, frame_h, window_w, window_h)
	id = util.tonumber (id)
	frame_x = util.tonumber (frame_x)
	frame_y = util.tonumber (frame_y)
	frame_w = util.tonumber (frame_w)
	frame_h = util.tonumber (frame_h)
	window_w = util.tonumber (window_w)
	window_h = util.tonumber (window_h)

	local bd = self.backdrops[id]

	local images = bd.images
	local frames = bd.frames
	
	-- Calculate suitable width and height
	
	local w = window_w
	local h = window_h
	
	if (bd.clamp_x < 4) then
		w = bd.images[1].width
	end
	if (bd.clamp_y < 4) then
		h = bd.images[1].height
	end
	
	if (bd.x_scroll == 1) then	
		w = bd.images[1].width
		if (bd.x_repeat == 1) then
			w = window_w + frame_w - math.floor (frame_w * bd.x_coef)
			if (bd.x_auto == 1) then
				w = w + bd.images[1].width
			end
		end
	end

	if (bd.y_scroll == 1) then
		h = bd.images[1].height
		if (bd.y_repeat == 1) then
			h = window_h + frame_h - math.floor (frame_h * bd.y_coef)
			if (bd.y_auto == 1) then
				h = h + bd.images[1].height
			end
		end
	end
	
	-- Clamping
	
	local clamp = 0
	if (bd.clamp_x == 1 or bd.clamp_x == 4) then
		if (bd.clamp_y == 1 or bd.clamp_y == 4) then
			clamp = 0 -- Top Left
		elseif (bd.clamp_y == 2 or bd.clamp_y == 5) then
			clamp = 6 -- Bottom Left
		end
	elseif (bd.clamp_x == 2 or bd.clamp_x == 5) then
		if (bd.clamp_y == 1 or bd.clamp_y == 4) then
			clamp = 2 -- Top Right
		elseif (bd.clamp_y == 2 or bd.clamp_y == 5) then
			clamp = 8 -- Bottom Right
		end
	end
	
	-- Double-edge clamping (resampling)
	
	local resample = 0
	local res_w = bd.images[1].width
	local res_h = bd.images[1].height
	
	if (bd.clamp_x == 3) then
		resample = 1
		res_w = frame_w - bd.x_offset - bd.x_offset
		w = frame_w - bd.x_offset
	elseif (bd.clamp_x == 6) then
		resample = 1
		res_w = window_w - bd.x_offset - bd.x_offset
		w = window_w - bd.x_offset
	end
	
	if (bd.clamp_y == 3) then
		resample = 1
		res_h = frame_h - bd.y_offset - bd.y_offset
		h = frame_h - bd.y_offset
	elseif (bd.clamp_y == 6) then
		resample = 1
		res_h = window_h - bd.y_offset - bd.y_offset
		h = window_h - bd.y_offset
	end
	
	w = w - bd.x_offset
	h = h - bd.y_offset
	
	MF_LoadBackdrop (id, bd.images[1].file, 0, 0, w, h, clamp, resample, res_w, res_h, bd.bg_type)
	
	for i = 2, #images do
		MF_AddBackdropFrame (id, i - 1, images[i].file)
	end
end

function Parallax:move (id, frame_x, frame_y, frame_w, frame_h, window_w, window_h)
	id = util.tonumber (id)
	frame_x = util.tonumber (frame_x)
	frame_y = util.tonumber (frame_y)
	frame_w = util.tonumber (frame_w)
	frame_h = util.tonumber (frame_h)
	window_w = util.tonumber (window_w)
	window_h = util.tonumber (window_h)
	
	local bd = self.backdrops[id]

	local base_x = frame_x
	local base_y = frame_y

	if (bd.x_scroll == 1) then
		if (bd.x_repeat == 1) then
			local auto_x = 0
			if (bd.x_auto ==1 ) then
				auto_x = (bd.auto_scroll * bd.x_speed) % bd.images[1].width
			end
			base_x = base_x * bd.x_coef - auto_x
		else
			local x_diff = bd.images[1].width - window_w
			local fx_diff = frame_w - window_w
			local x_factor = 1.0 * frame_x / fx_diff
			base_x = frame_x - (x_diff * x_factor)
		end
	else
		if (bd.clamp_x == 1 or bd.clamp_x == 3) then
			base_x = bd.x_offset
		elseif (bd.clamp_x == 2) then
			base_x = frame_w - bd.images[1].width - bd.x_offset
		elseif (bd.clamp_x == 4 or bd.clamp_x == 6) then
			base_x = frame_x + bd.x_offset
		elseif (bd.clamp_x == 5) then
			base_x = frame_x + window_w - bd.images[1].width - bd.x_offset
		end
	end
	
	if (bd.y_scroll == 1) then
		if (bd.y_repeat == 1) then
			local auto_y = 0
			if (bd.y_auto ==1 ) then
				auto_y = (bd.auto_scroll * bd.y_speed) % bd.images[1].height
			end
			base_y = base_y * bd.y_coef - auto_y
		else
			local y_diff = bd.images[1].height - window_h
			local fy_diff = frame_h - window_h
			local y_factor = 1.0 * frame_y / fy_diff
			base_y = frame_y - (y_diff * y_factor)
		end
	else
		if (bd.clamp_y == 1 or bd.clamp_y == 3) then
			base_y = bd.y_offset
		elseif (bd.clamp_y == 2) then
			base_y = frame_h - bd.images[1].height - bd.y_offset
		elseif (bd.clamp_y == 4 or bd.clamp_y == 6) then
			base_y = frame_y + bd.y_offset
		elseif (bd.clamp_y == 5) then
			base_y = frame_y + window_h - bd.images[1].height - bd.y_offset
		end
	end
	
	self.backdrops[id].auto_scroll = bd.auto_scroll + 1
	
	local anim_speed = 8
	self.backdrops[id].cur_frame = bd.cur_frame + 1
	
	local cur_frame = bd.cur_frame / anim_speed
	if (cur_frame >= #bd.frames) then
		self.backdrops[id].cur_frame = 0
		cur_frame = 0
	end

	MF_MoveBackdrop (id, base_x, base_y, cur_frame)
end

function Parallax:zList ()
	local sortList = {}
	for k,v in pairs(self.backdrops) do
		table.insert (sortList, v)
	end
	table.sort (sortList, function(a,b) return a.z_order < b.z_order end)
	return sortList
end

function Parallax:import (data, ver)
	local offPara = util.unu32(string.sub(data, 1, 4)) - 8
	local numPara = util.unu32(string.sub(data, 5, 8))
	local szPara = util.unu16(string.sub(data, 9, 10))
	local offImage = util.unu32(string.sub(data, 11, 14)) - 8
	local numImage = util.unu32(string.sub(data, 15, 18))
	local szImage = util.unu16(string.sub(data, 19, 20))
	local offAnim = util.unu32(string.sub(data, 21, 24)) - 8
	local numAnim = util.unu32(string.sub(data, 25, 28))
	local szAnim = util.unu16(string.sub(data, 29, 30))
	
	self.backdrops = {}
	self.lastIndex = 0
	
	if (ver < 1) then
		return
	end
		
	for iPara = 1, numPara do
		local addrPara = offPara + (iPara - 1) * szPara
		local id = util.unu32(string.sub(data, addrPara+1, addrPara+4))
		local name = util.unfixstr(string.sub(data, addrPara+5, addrPara+36))

		local anim_idx = util.unu32(string.sub(data, addrPara+37, addrPara+40))
		local anim_num = util.unu16(string.sub(data, addrPara+41, addrPara+42))
		local anim_speed = util.unu16(string.sub(data, addrPara+43, addrPara+44))
		local bg_type = util.unu8(string.sub(data, addrPara+45, addrPara+45))
		local bg_start = util.unu8(string.sub(data, addrPara+46, addrPara+46))
		
		local x_scroll = util.unu8(string.sub(data, addrPara+47, addrPara+47))
		local y_scroll = util.unu8(string.sub(data, addrPara+48, addrPara+48))
		local x_repeat = util.unu8(string.sub(data, addrPara+49, addrPara+49))
		local y_repeat = util.unu8(string.sub(data, addrPara+50, addrPara+50))
		local x_auto = util.unu8(string.sub(data, addrPara+51, addrPara+51))
		local y_auto = util.unu8(string.sub(data, addrPara+52, addrPara+52))
		local z_order = util.unu16(string.sub(data, addrPara+53, addrPara+54))
		local x_coef = util.unfs8(string.sub(data, addrPara+55, addrPara+62))
		local y_coef = util.unfs8(string.sub(data, addrPara+63, addrPara+70))
		local x_speed = util.unfs8(string.sub(data, addrPara+71, addrPara+78))
		local y_speed = util.unfs8(string.sub(data, addrPara+79, addrPara+86))
		local clamp_x = util.unu8(string.sub(data, addrPara+87, addrPara+87))
		local clamp_y = util.unu8(string.sub(data, addrPara+88, addrPara+88))
		local x_offset = util.unu32(string.sub(data, addrPara+89, addrPara+92))
		local y_offset = util.unu32(string.sub(data, addrPara+93, addrPara+96))
		
		id = self:add (0, name, z_order, bg_type, bg_start, anim_speed, clamp_x, clamp_y, x_offset, y_offset)
		self:updateScrolling (id, x_scroll, x_repeat, x_auto, x_coef, x_speed, y_scroll, y_repeat, y_auto, y_coef, y_speed)

		for iAnim = 1, anim_num do
			local addrAnim = offAnim + (anim_idx + iAnim - 1) * szAnim
			local image_idx = util.unu32(string.sub(data, addrAnim+1, addrAnim+4))
			
			local addrImage = offImage + (image_idx - 1) * szImage
			local image_file = util.unfixstr(string.sub(data, addrImage+1, addrImage+64))
			local image_w = util.unu16(string.sub(data, addrImage+65, addrImage+66))
			local image_h = util.unu16(string.sub(data, addrImage+67, addrImage+68))

			self:addFrame (id, image_file, image_w, image_h)
			--self:addFrame (id, "cave1_2x_2.png", image_w, image_h)
			--self:addFrame (id, "cave1_2x_3.png", image_w, image_h)
			--self:addFrame (id, "cave1_2x_4.png", image_w, image_h)
		end
	end
end
	
function Parallax:export ()
	local data = ""
	local dataPara = ""
	local dataImage = ""
	local dataFrame = ""
	local idxPara = 0
	local idxImage = 0
	local idxFrame = 0
	
	for k,v in pairs( self.backdrops ) do
		dataPara = dataPara .. util.u32(v.id) .. util.fixstr(v.name, 32) .. util.u32(idxFrame) .. 
				   util.u16(#v.frames) .. util.u16(v.anim_speed) .. util.u8(v.bg_type) .. util.u8(v.bg_start) ..
				   util.u8(v.x_scroll) .. util.u8(v.y_scroll) .. util.u8(v.x_repeat) .. util.u8(v.y_repeat) .. 
				   util.u8(v.x_auto) .. util.u8(v.y_auto) .. util.u16(1) .. util.fs8(v.x_coef) .. 
				   util.fs8(v.y_coef) .. util.fs8(v.x_speed) .. util.fs8(v.y_speed) .. util.u8(v.clamp_x) ..
				   util.u8(v.clamp_y) .. util.u32(v.x_offset) .. util.u32(v.y_offset)
		idxPara = idxPara + 1
		
		for i = 1, #v.frames do
			dataFrame = dataFrame .. util.u32(idxImage + v.frames[i])
			idxFrame = idxFrame + 1
		end
		
		for i = 1, #v.images do
			dataImage = dataImage .. util.fixstr(v.images[i].file, 64) .. util.u16(v.images[i].width) ..
						util.u16(v.images[i].height)
			idxImage = idxImage + 1
		end

	end
	
	local offsetPara = self.filep.szBlock
	local offsetImage = offsetPara + self.filep.szPara * idxPara
	local offsetFrame = offsetImage + self.filep.szImage * idxImage
	
	data = self.filep.mime .. util.u16(self.filep.ver) .. util.u16(self.filep.szBlock) .. 
		   util.u32(self.filep.szBlock) .. util.u32(idxPara) .. util.u16(self.filep.szPara) ..
		   util.u32(offsetImage) .. util.u32(idxImage) .. util.u16(self.filep.szImage) ..
		   util.u32(offsetFrame) .. util.u32(idxFrame) .. util.u16(self.filep.szFrame)
	return data .. dataPara .. dataImage .. dataFrame
end
	
function Parallax:reset ()
	self.backdrops = {}
	self.lastIndex = 0
end