local util = require "./lua/util"
local undo = require "./lua/undo"

local string = string
local table = table

local pcall = pcall
local setmetatable = setmetatable
local pairs = pairs
local ipairs = ipairs
local tonumber = tonumber
local DoCall = DoCall
local error = error

local MF_Error = MF_Error
local MF_BackgroundClearTile = MF_BackgroundClearTile
local MF_BackgroundDrawTile = MF_BackgroundDrawTile
local MF_CollisionClearTile = MF_CollisionClearTile
local MF_CollisionDrawTile = MF_CollisionDrawTile
local MF_ForegroundClearTile = MF_ForegroundClearTile
local MF_ForegroundDrawTile = MF_ForegroundDrawTile

module (...)

-- ----------------------------------------------------------------------------

function TileClass ()
	return Tile:new()
end

Tile = {
	filep = { mime = "", ver = 0, szBlock = 0, szTile = 0, szLayer = 0 },
	map = {},
	slice = {},
	slicep = { x1 = 0, y1 = 0, x2 = 0, y2 = 0, offx = 0, offy = 0 },
	fl_slice = {},
	fl_clipboard = {},
	chip_width = 0,
	chip_height = 0,
	undoStack = undo.newUndoStack(),
	undoAvail = false,
	redoAvail = false,
	ttc = nil,
	frame = nil,
	last_signature = "",
	last_block_signature = "",
}

function Tile:new (obj)
	obj = obj or {}
	setmetatable (obj, self)
	self.__index = self
	return obj
end

-- Primitives -------------------------
	
function Tile:add ( dx, dy, sx, sy, ... )
	if (dx < 0 or dy < 0) then
		MF_Error ("Tile.add: Coordinates out of bounds [" .. dx .. ", " .. dy .. "]")
		return
	end
	
	local d = util.makeid( dx, dy )
	local s = util.makeid( sx, sy )
	local signature = d .. "_" .. s
	
	if (self.last_signature == signature) then
		return
	end
	
	self.last_signature = signature
	
	if (self.map[d] == nil) then
		self.map[d] = {
			dx = util.tonumber (dx),
			dy = util.tonumber (dy),
			cnt = 0,
			depth = 0,
			tile = {},
		}
	end

	if ( self.map[d].tile[s] == nil ) then
		self.map[d].cnt = self.map[d].cnt + 1
		self.map[d].depth = self.map[d].depth + 1
		self.map[d].tile[s] = self:_layer( sx, sy, ... )
		self.map[d].tile[s].depth = self.map[d].depth
		setmetatable( self.map[d].tile[s], setmt )
	else
		self.map[d].depth = self.map[d].depth + 1
		self.map[d].tile[s].depth = self.map[d].depth
	end
	self:_draw( dx, dy, sx, sy )
end

function Tile:clear ( dx, dy )
	local d = util.makeid( dx, dy )
	if ( self.map[d] ~= nil ) then
		-- Log undo action
		--local udata = { dx = dx, dy = dy, v = util.clone (self.map[d]) }
		--self:logUndo ("clear", udata, 0)
		
		self.signature_reset()
		
		self.map[d] = nil
		self:_clear( dx, dy )
	end
end

function Tile:redrawAt ( dx, dy, nx, ny )
	local d = util.makeid( dx, dy )
	
	if ( self.map[d] ~= nil ) then
		local layerList = util.depthsort( self.map[d].tile )
		for k1,v1 in ipairs(layerList) do
			self:_draw( nx, ny, v1.sx, v1.sy )
		end
	end
end

-- Compound ---------------------------

function Tile:addBlock (dx, dy, sx, sy, sw, sh, ...)
	local udata = { }
	
	-- Return if we already drew this block
	local d = util.makeid( dx, dy )
	local s = util.makeid( sx, sy )
	local w = util.makeid( sw, sh )
	local signature = d .. "_" .. s .. "_" .. w
	
	if (self.last_block_signature == signature) then
		return
	end
	
	self.last_block_signature = signature
	
	for ix = 1, sw do
		for iy = 1, sh do
			-- Create undo data structures
			local d_id = util.makeid (dx + ix - 1, dy + iy - 1)
			table.insert (udata, { dx = dx + ix - 1, dy = dy + iy - 1, v = util.clone (self.map[d_id]) } )
			
			-- Add single component tile
			self:add (dx + ix - 1, dy + iy - 1, sx + ix - 1, sy + iy - 1, ...)
		end
	end
	
	-- Log undo action
	self:logUndo ("addMulti", udata, 0)
end

function Tile:fillBlock (dx, dy, sx, sy, dw, dh, sw, sh, ...)
	local udata = { }
	
	for ix = 1, dw do
		for iy = 1, dh do
			-- Create undo data structures
			local d_id = util.makeid (dx + ix - 1, dy + iy - 1)
			table.insert (udata, { dx = dx + ix - 1, dy = dy + iy - 1, v = util.clone (self.map[d_id]) } )
			
			-- Add single component tile
			self:add (dx + ix - 1, dy + iy - 1, sx + ( (ix - 1) % sw), sy + ( (iy - 1) % sh), ...)
		end
	end
	
	-- Log undo action
	self:logUndo ("addMulti", udata, 0)
end

-- Floating Selection -----------------

function Tile:createFloat (x1, y1, x2, y2, cx)
	local udata = { }
	local floating = {
		orig_x = x1,
		orig_y = y1,
		x = x1,
		y = y1,
		width = x2 - x1,
		height = y2 - y1,
		data = {}
	}
	
	for ix = x1,x2 do
		for iy = y1,y2 do
			-- Create undo data structures
			local d = util.makeid (ix, iy)
			table.insert (udata, { dx = ix, dy = iy, v = util.clone (self.map[d]) } )
			
			-- Add to floating selection
			local f = util.makeid (ix-x1, iy-y1)
			floating.data[f] = util.clone (self.map[d])
			if (cx == 'cut') then
				self.map[d] = nil
			end
		end
	end
	
	-- Conditionally log undo action
	if (cx == 'cut') then
		self:logUndo ("clearMulti", udata, 0)
	end

	return floating
end

function Tile:pasteFloat (fl)
	local udata = { }
	
	for ix = fl.x,fl.x+fl.width do
		for iy = fl.y,fl.y+fl.height do
			-- Create undo data structures
			local d = util.makeid (ix, iy)
			table.insert (udata, { dx = ix, dy = iy, v = util.clone (self.map[d]) } )
			
			-- Merge component tiles
			local sd = util.makeid (ix-fl.x, iy-fl.y)
			if (fl.data[sd] ~= nil) then
				self.map[d] = self:merge_tiles (self.map[d], fl.data[sd])
				self.map[d].dx = self.map[d].dx + (fl.x - fl.orig_x)
				self.map[d].dy = self.map[d].dy + (fl.y - fl.orig_y)
			end
		end
	end
	
	-- Conditionally log undo action
	if (fl.x ~= fl.orig_x or fl.y ~= fl.orig_y) then
		self:logUndo ("addMulti", udata, 0)
	else
		self:undo()
		self.undoStack:Redo()
		self.redoAvail = self.undoStack:RedoAvailable()
	end
end

function Tile:redrawFloatAt (fl, fx, fy, nx, ny)
	if (fl == nil) then
		return
	end
	
	local d = util.makeid (fx, fy)
	if (fl.data[d] ~= nil) then
		local layerlist = util.depthsort (fl.data[d].tile)
		for k1,v1 in ipairs (layerlist) do
			self:_draw (nx, ny, v1.sx, v1.sy)
		end
	end
end

-- Utility ----------------------------

function Tile:signature_reset ()
	self.last_signature = ""
	self.last_block_signature = ""
end

function Tile:merge_tiles (t1, t2)
	if (t1 == nil) then
		return util.clone (t2)
	elseif (t2 == nil) then
		return util.clone (t1)
	end
	
	local t3 = {
		dx = t2.dx,
		dy = t2.dy,
		cnt = 0,
		depth = 0,
		tile = {},
	}
	
	local layerList = util.depthsort (t1.tile)
	for k,v in ipairs (layerList) do
		local sid = util.makeid (v.sx, v.sy)
		t3.depth = t3.depth + 1
		t3.cnt = t3.cnt + 1
		t3.tile[sid] = util.clone (v)
		t3.tile[sid].depth = t3.depth
	end
	
	local layerList = util.depthsort (t2.tile)
	for k,v in ipairs (layerList) do
		local sid = util.makeid (v.sx, v.sy)
		t3.depth = t3.depth + 1
		if (t3.tile[sid] == nil) then
			t3.cnt = t3.cnt + 1
			t3.tile[sid] = util.clone (v)
		end
		t3.tile[sid].depth = t3.depth
	end

	return t3
end

-- Undo / Redo ------------------------

function Tile:logUndo (_name, _data, _type)
	local udata = {
		name = _name,
		data = _data
	}

	if (_type == 0) then
		self.undoStack:logData (udata)
		self.undoAvail = true
		self.redoAvail = false
	elseif (_type == 1) then
		self.undoStack:logUndo (udata)
		self.undoAvail = true
	elseif (_type == 2) then
		self.undoStack:logRedo (udata)
		self.redoAvail = true
	end
end

function Tile:undo ()
	local udata = self.undoStack:Undo()
	if (udata == nil) then
		return false
	end
	
	if (udata.name == "add") then
		self:_undo_add_tile (udata, 2)
	elseif (udata.name == "addMulti") then
		self:_undo_add_multi_tile (udata, 2)
	elseif (udata.name == "clear") then
		self:_undo_clear_tile (udata, 2)
	elseif (udata.name == "clearMulti") then
		self:_undo_clear_multi_tile (udata, 2)
	end
	
	self.undoAvail = self.undoStack:UndoAvailable()
	self.redoAvail = self.undoStack:RedoAvailable()
end

function Tile:redo ()
	local rdata = self.undoStack:Redo()
	if (rdata == nil) then
		return false
	end

	if (rdata.name == "add") then
		self:_undo_add_tile (rdata, 1)
	elseif (rdata.name == "addMulti") then
		self:_undo_add_multi_tile (rdata, 1)
	elseif (rdata.name == "clear") then
		self:_undo_clear_tile (rdata, 1)
	elseif (rdata.name == "clearMulti") then
		self:_undo_clear_multi_tile (rdata, 1)
	end
	
	self.undoAvail = self.undoStack:UndoAvailable()
	self.redoAvail = self.undoStack:RedoAvailable()
end

function Tile:_undo_add_tile (udata, _type)
	local d = util.makeid (udata.data.dx, udata.data.dy)
		
	-- Log redo action
	local rdata = { dx = udata.data.dx, dy = udata.data.dy, v = util.clone (self.map[d]) }
	self:logUndo ("clear", rdata, _type)
	
	self.map[d] = util.clone (udata.data.v)
	self:_clear (udata.data.dx, udata.data.dy)
	self:redrawAt (udata.data.dx, udata.data.dy, udata.data.dx, udata.data.dy)
end

function Tile:_undo_add_multi_tile (udata, _type)
	local rdata = { }
	
	for k,v in pairs (udata.data) do
		-- Create redo data structures
		local d_id = util.makeid (v.dx, v.dy)
		table.insert (rdata, { dx = v.dx, dy = v.dy, v = util.clone (self.map[d_id]) } )
		
		-- Perform undo on single tile
		self.map[d_id] = util.clone (v.v)
		self:_clear (v.dx, v.dy)
		self:redrawAt (v.dx, v.dy, v.dx, v.dy)
	end
	
	-- Log redo action
	self:logUndo ("clearMulti", rdata, _type)
end
	


function Tile:_undo_clear_tile (udata, _type)
	local d = util.makeid (udata.data.dx, udata.data.dy)

	-- Log redo action
	local rdata = { dx = udata.data.dx, dy = udata.data.dy, v = nil }
	self:logUndo ("add", rdata, _type)
	
	self.map[d] = util.clone (udata.data.v)
	self:_clear (udata.data.dx, udata.data.dy)
	self:redrawAt (udata.data.dx, udata.data.dy, udata.data.dx, udata.data.dy)
end

function Tile:_undo_clear_multi_tile (udata, _type)
	local rdata = { }
	
	for k,v in pairs (udata.data) do
		-- Create redo data structures
		local d_id = util.makeid (v.dx, v.dy)
		table.insert (rdata, { dx = v.dx, dy = v.dy, v = util.clone (self.map[d_id]) } )
		
		-- Perform undo on single tile
		self.map[d_id] = util.clone (v.v)
		self:_clear (v.dx, v.dy)
		self:redrawAt (v.dx, v.dy, v.dx, v.dy)
	end
	
	-- Log redo action
	self:logUndo ("addMulti", rdata, _type)
end

-- Import / Export --------------------
	
function Tile:import ( data, ver )
	local offTile = util.unu32(string.sub(data, 1, 4)) - 8
	local offLayer = util.unu32(string.sub(data, 5, 8)) - 8
	local numTile = util.unu32(string.sub(data, 9, 12))
	local numLayer = util.unu32(string.sub(data, 13,16))
	local szTile = util.unu16(string.sub(data, 17, 18))
	local szLayer = util.unu16(string.sub(data, 19, 20))

	for iTile = 1, numTile do
		local addrTile = offTile + (iTile - 1) * szTile
		local idxLayer = util.unu32(string.sub(data, addrTile+1, addrTile+4))
		local instLayer = util.unu16(string.sub(data, addrTile+5, addrTile+6))
		local dx = util.unu16(string.sub(data, addrTile+7, addrTile+8))
		local dy = util.unu16(string.sub(data, addrTile+9, addrTile+10))
		
		for iLayer = 1, instLayer do
			local addrLayer = offLayer + (idxLayer + iLayer - 1) * szLayer
			local sx = util.unu16(string.sub(data, addrLayer+1, addrLayer+2))
			local sy = util.unu16(string.sub(data, addrLayer+3, addrLayer+4))
			local lparams = self:_importLayer( data, addrLayer )
			self:add( dx, dy, sx, sy, lparams )
		end
	end
	
	self.undoStack = undo.newUndoStack()
end
	
function Tile:export ()
	local data = ""
	local dataTile = ""
	local dataLayer = ""
	local idxTile = 0
	local idxLayer = 0
	
	local stat, err = pcall (function ()
		for kt,vt in pairs( self.map ) do
			dataTile = dataTile .. util.u32(idxLayer) .. util.u16(vt.cnt) .. util.u16(vt.dx) .. util.u16(vt.dy)
			local tmpLayer = {}
			local maxdepth = 0
			for kl,vl in pairs( vt.tile ) do
				tmpLayer[vl.depth] = vl
				if (vl.depth > maxdepth) then maxdepth = vl.depth end
			end
			for i=1, maxdepth do
				if (tmpLayer[i] ~= nil) then
					dataLayer = dataLayer .. util.u16(tmpLayer[i].sx) .. util.u16(tmpLayer[i].sy)
					dataLayer = dataLayer .. self:_exportLayer(tmpLayer[i])
					idxLayer = idxLayer + 1
				end
			end
			idxTile = idxTile + 1
		end
		
		data = self.filep.mime .. util.u16(self.filep.ver) .. util.u16(self.filep.szBlock) ..
		   util.u32(self.filep.szBlock) .. util.u32(self.filep.szBlock + self.filep.szTile*idxTile) ..
		   util.u32(idxTile) .. util.u32(idxLayer) .. util.u16(self.filep.szTile) .. util.u16(self.filep.szLayer)
	end )
	
	if (stat == false) then
		MF_Error ("Tile.export: Processing Error\n" .. err)
	end

	return data .. dataTile .. dataLayer
end
	
function Tile:reset ()
	self.map = {}
	self.slice = {}
end

-- ----------------------------------------------------------------------------

function CollisionClass ()
	return CollisionTile:new()
end

CollisionTile = Tile:new {
	filep = { mime = "COLL", ver = 0, szBlock = 28, szTile = 10, szLayer = 6 },
	map = {},
	slice = {},
	slicep = { x1 = 0, y1 = 0, x2 = 0, y2 = 0, offx = 0, offy = 0 },
	chip_width = 0,
	chip_height = 0,
	undoStack = undo.newUndoStack(),
	undoAvail = false,
	redoAvail = false,
}

function CollisionTile:_layer ( sx, sy, t )
	return {
		sx = util.tonumber (sx),
		sy = util.tonumber (sy),
		t = util.tonumber (t),
	}
end

function CollisionTile:_clear ( dx, dy )
	if (self.ttc ~= nil) then
		dx, dy = self.ttc (dx, dy)
	end
	--DoCall( "CollisionClearTile", dx, dy )
	MF_CollisionClearTile (dx, dy)
end

function CollisionTile:_draw ( nx, ny, sx, sy )
	if (self.ttc ~= nil) then
		nx, ny = self.ttc (nx, ny)
	end
	--DoCall( "CollisionDrawTile", nx, ny, sx, sy )
	MF_CollisionDrawTile (nx, ny, sx, sy, self.frame.tileWidth, self.frame.tileHeight)
end
	
function CollisionTile:_importLayer (data, addr)
	local t = util.unu16(string.sub(data, addr+5, addr+6))
	return t
end
	
function CollisionTile:_exportLayer (layer)
	return util.u16(layer.t)
end

-- ----------------------------------------------------------------------------

function BackgroundClass ()
	return BackgroundTile:new()
end

BackgroundTile = Tile:new {
	filep = { mime = "BACK", ver = 0, szBlock = 28, szTile = 10, szLayer = 4 },
	map = {},
	slice = {},
	slicep = { x1 = 0, y1 = 0, x2 = 0, y2 = 0, offx = 0, offy = 0 },
	chip_width = 0,
	chip_height = 0,
	undoStack = undo.newUndoStack(),
	undoAvail = false,
	redoAvail = false,
}

function BackgroundTile:_layer ( sx, sy )
	return {
		sx = util.tonumber (sx),
		sy = util.tonumber (sy),
	}
end
	
function BackgroundTile:_clear ( dx, dy )
	if (self.ttc ~= nil) then
		dx, dy = self.ttc (dx, dy)
	end
	--DoCall( "BackgroundClearTile", dx, dy )
	MF_BackgroundClearTile (dx, dy)
end
	
function BackgroundTile:_draw ( nx, ny, sx, sy )
	if (self.ttc ~= nil) then
		nx, ny = self.ttc (nx, ny)
	end
	--DoCall( "BackgroundDrawTile", nx, ny, sx, sy )
	MF_BackgroundDrawTile (nx, ny, sx, sy, self.frame.tileWidth, self.frame.tileHeight)
end
	
function BackgroundTile:_importLayer (data, addr)
	return {}
end
	
function BackgroundTile:_exportLayer (layer)
	return ""
end

-- ----------------------------------------------------------------------------

function ForegroundClass ()
	return ForegroundTile:new()
end

ForegroundTile = Tile:new {
	filep = { mime = "FORE", ver = 0, szBlock = 28, szTile = 10, szLayer = 4 },
	map = {},
	slice = {},
	slicep = { x1 = 0, y1 = 0, x2 = 0, y2 = 0, offx = 0, offy = 0 },
	chip_width = 0,
	chip_height = 0,
	undoStack = undo.newUndoStack(),
	undoAvail = false,
	redoAvail = false,
}

function ForegroundTile:_layer ( sx, sy )
	return {
		sx = util.tonumber (sx),
		sy = util.tonumber (sy),
	}
end
	
function ForegroundTile:_clear ( dx, dy )
	if (self.ttc ~= nil) then
		dx, dy = self.ttc (dx, dy)
	end
	--DoCall( "ForegroundClearTile", dx, dy )
	MF_ForegroundClearTile (dx, dy)
end
	
function ForegroundTile:_draw ( nx, ny, sx, sy )
	if (self.ttc ~= nil) then
		nx, ny = self.ttc (nx, ny)
	end
	--DoCall( "ForegroundDrawTile", nx, ny, sx, sy )
	MF_ForegroundDrawTile (nx, ny, sx, sy, self.frame.tileWidth, self.frame.tileHeight)
end
	
function ForegroundTile:_importLayer (data, addr)
	return {}
end
	
function ForegroundTile:_exportLayer (layer)
	return ""
end