local math = math
local table = table
local string = string

local setmetatable = setmetatable
local pairs = pairs
local next = next
local type = type
local error = error
local _tostring = tostring
local _tonumber = tonumber

module (...)

-- ----------------------------------------------------------------------------

tostring = _tostring
tonumber = _tonumber

setmt = {
	__tostring = function (e)
		str = "{\n"
		for k,v in pairs(e) do
			str = str .. k .. " => " .. tostring(v) .. "\n"
		end
		str = str .. "}\n"
		return str
	end
}

function depthsort( layers )
	sortLayer = {}
	for k,v in pairs(layers) do
		table.insert( sortLayer, v )
	end
	table.sort( sortLayer, function(a,b) return a.depth < b.depth end )
	setmetatable( sortLayer, setmt )
	return sortLayer
end

function paramsort( ar, p )
	sortar = {}
	for k,v in pairs(ar) do
		table.insert( sortar, v )
	end
	table.sort( sortar, function(a,b) return a[p] < b[p] end )
	setmetatable( ar, setmt )
	return sortar
end

function makeid( a, b )
	return a .. "_" .. b
end

function u32( i )
	i = tonumber (i)
	if (i == nil) then
		error ("util.u32: nil argument passed", 2)
	end
	if (i < 0) then
		error ("util.u32: integer underflow [" .. tostring (i) .. "]", 2)
	end
	local f1 = math.floor(i / 16777216)
	local r1 = i % 16777216
	local f2 = math.floor(r1 / 65536)
	local r2 = r1 % 65536
	local f3 = math.floor(r2 / 256)
	local f4 = r2 % 256
	if (f1 > 255) then
		error ("util.u32: integer overflow [" .. tostring (i) .. "]", 2)
	end
	return string.char(f4) .. string.char(f3) .. string.char(f2) .. string.char(f1)
end

function s32 (i)
	i = tonumber (i)
	if (math.abs (i) > 2147483647) then
		error ("util.s32: integer overflow [" .. tostring (i) .. "]", 2)
	end
	if (i < 0) then
		i = i + 4294967296
	end
	local f1 = math.floor(i / 16777216)
	local r1 = i % 16777216
	local f2 = math.floor(r1 / 65536)
	local r2 = r1 % 65536
	local f3 = math.floor(r2 / 256)
	local f4 = r2 % 256
	return string.char(f4) .. string.char(f3) .. string.char(f2) .. string.char(f1)
end

function u16( i )
	i = tonumber (i)
	if (i == nil) then
		error ("util.u16: nil argument passed", 2)
	end
	if (i < 0) then
		error ("util.u16: integer underflow [" .. tostring (i) .. "]", 2)
	end
	local f1 = math.floor(i / 256)
	local f2 = i % 256
	if (f1 > 255) then
		error ("util.u16: integer overflow [" .. tostring (i) .. "]", 2)
	end
	return string.char(f2) .. string.char(f1)
end

function s16 (i)
	i = tonumber (i)
	if (math.abs (i) > 32767) then
		error ("util.s16: integer overflow [" .. tostring (i) .. "]", 2)
	end
	if (i < 0) then
		i = i + 65536
	end
	local f1 = math.floor(i / 256)
	local f2 = i % 256
	return string.char(f2) .. string.char(f1)
end

function u8 ( i )
	i = tonumber (i)
	if (i == nil) then
		error ("util.u8: nil argument passed", 2)
	end
	if (i < 0) then
		error ("util.u8: integer underflow [" .. tostring (i) .. "]", 2)
	end
	local f1 = i % 256
	if (i > 255) then
		error ("util.u8: integer overflow [" .. tostring (i) .. "]", 2)
	end
	return string.char(f1)
end

function s8 (i)
	i = tonumber (i)
	if (math.abs (i) > 127) then
		error ("util.s8: integer overflow [" .. tostring (i) .. "]", 2)
	end
	if (i < 0) then
		i = i + 256
	end
	return string.char(i)
end

function unu32( str )
	if (str == nil or str == "") then
		error ("util.unu32: empty argument passed", 2)
	end
	local i = 0
	i = i + string.byte(string.sub(str, 1, 1))
	i = i + string.byte(string.sub(str, 2, 2)) * 256
	i = i + string.byte(string.sub(str, 3, 3)) * 65536
	i = i + string.byte(string.sub(str, 4, 4)) * 16777216
	return i
end

function uns32( str )
	local i = 0
	i = i + string.byte(string.sub(str, 1, 1))
	i = i + string.byte(string.sub(str, 2, 2)) * 256
	i = i + string.byte(string.sub(str, 3, 3)) * 65536
	i = i + string.byte(string.sub(str, 4, 4)) * 16777216
	if (i > 2147483647) then
		i = i - 4294967296
	end
	return i
end

function unu16( str )
	local i = 0
	i = i + string.byte(string.sub(str, 1, 1))
	i = i + string.byte(string.sub(str, 2, 2)) * 256
	return i
end

function uns16( str )
	local i = 0
	i = i + string.byte(string.sub(str, 1, 1))
	i = i + string.byte(string.sub(str, 2, 2)) * 256
	if (i > 32767) then
		i = i - 65536
	end
	return i
end

function unu8( str )
	local i = 0
	i = i + string.byte(string.sub(str, 1, 1))
	return i
end

function uns8( str )
	local i = 0
	i = i + string.byte(string.sub(str, 1, 1))
	if (i > 127) then
		i = i - 256
	end
	return i
end

function fixstr( str, len )
	str = tostring (str)
	local data = string.sub(str, 1, math.min(len, #str))
	for i = #str+1,len do
		data = data .. string.char(0)
	end
	return data
end

function unfixstr( str )
	local idx = string.find( str, string.char(0), 1, true )
	if (idx == nil) then
		return str
	end
	return string.sub( str, 1, idx-1 )
end

function fs8( str )
	return fixstr( str, 8 )
end

function unfs8( str )
	local ustr = unfixstr( str )
	return tonumber( ustr )
end

function basename( str )
	if (str == nil) then
		error ("Nil string passed to util.basename", 2)
	end
	
	idx1, idx2, mt = string.find( str, "[/\\]" )
	if ( idx1 ~= nil ) then
		return basename( string.sub( str, idx2+1, #str ) )
	end
	return str
end

function clone (tab)
	if (tab == nil) then
		return nil
	end
	
	local nt = {}
	for k,v in pairs (tab) do
		if type (v) == "table" then
			v = clone (v)
		end
		nt[k] = v
	end
	return nt
end

function tablesize (tab)
	if (tab == nil) then
		return -1
	end
	
	local count = 0
	for v in pairs (tab) do
		count = count + 1
	end
	
	return count
end

function str_join (delimiter, list)
	local len = table.getn(list)
	if len == 0 then 
		return "" 
	end
	
	local string = list[1]
	for i = 2, len do 
		string = string .. delimiter .. list[i] 
	end
	
	return string
end

function str_split (delim, str, maxNb)
    -- Eliminate bad cases...
    if string.find(str, delim) == nil then
        return { str }
    end
    if maxNb == nil or maxNb < 1 then
        maxNb = 0    -- No limit
    end
    local result = {}
    local pat = "(.-)" .. delim .. "()"
    local nb = 0
    local lastPos
    for part, pos in string.gfind(str, pat) do
        nb = nb + 1
        result[nb] = part
        lastPos = pos
        if nb == maxNb then break end
    end
    -- Handle the last field
    if nb ~= maxNb then
        result[nb + 1] = string.sub(str, lastPos)
    end
    return result
end

